#!/usr/bin/env python3
# Impacket - Collection of Python classes for working with network protocols.
#
# SECUREAUTH LABS. Copyright (C) 2021 SecureAuth Corporation. All rights reserved.
#
# This software is provided under a slightly modified version
# of the Apache Software License. See the accompanying LICENSE file
# for more information.
#
# Description:
#   Python script to read and manage the Discretionary Access Control List of an object
#
# Authors:
#   Charlie BROMBERG (@_nwodtuhs)
#   Guillaume DAUMAS (@BlWasp_)
#   Lucien DOUSTALY (@Wlayzz)
#

import argparse
import binascii
import codecs
import json
import logging
import os
import sys
import traceback
import datetime

import ldap3
import ssl
import ldapdomaindump
from binascii import unhexlify
from enum import Enum
from ldap3.protocol.formatters.formatters import format_sid

from impacket import version
from impacket.examples import logger, utils
from impacket.ldap import ldaptypes
from powerview.utils.constants import (
    SCHEMA_OBJECTS,
    EXTENDED_RIGHTS
)
from impacket.smbconnection import SMBConnection
from impacket.spnego import SPNEGO_NegTokenInit, TypesMech
from ldap3.utils.conv import escape_filter_chars
from ldap3.protocol.microsoft import security_descriptor_control
from impacket.uuid import string_to_bin, bin_to_string

from powerview.utils.constants import WELL_KNOWN_SIDS

OBJECT_TYPES_GUID = {}
OBJECT_TYPES_GUID.update(SCHEMA_OBJECTS)
OBJECT_TYPES_GUID.update(EXTENDED_RIGHTS)

# GUID rights enum
# GUID thats permits to identify extended rights in an ACE
# https://docs.microsoft.com/en-us/windows/win32/adschema/a-rightsguid
class RIGHTS_GUID(Enum):
    WriteMembers = "bf9679c0-0de6-11d0-a285-00aa003049e2"
    ResetPassword = "00299570-246d-11d0-a768-00aa006e0529"
    DS_Replication_Get_Changes = "1131f6aa-9c07-11d1-f79f-00c04fc2dcd2"
    DS_Replication_Get_Changes_All = "1131f6ad-9c07-11d1-f79f-00c04fc2dcd2"


# ACE flags enum
# New ACE at the end of SACL for inheritance and access return system-audit
# https://docs.microsoft.com/en-us/windows/win32/api/securitybaseapi/nf-securitybaseapi-addauditaccessobjectace
class ACE_FLAGS(Enum):
    CONTAINER_INHERIT_ACE = ldaptypes.ACE.CONTAINER_INHERIT_ACE
    FAILED_ACCESS_ACE_FLAG = ldaptypes.ACE.FAILED_ACCESS_ACE_FLAG
    INHERIT_ONLY_ACE = ldaptypes.ACE.INHERIT_ONLY_ACE
    INHERITED_ACE = ldaptypes.ACE.INHERITED_ACE
    NO_PROPAGATE_INHERIT_ACE = ldaptypes.ACE.NO_PROPAGATE_INHERIT_ACE
    OBJECT_INHERIT_ACE = ldaptypes.ACE.OBJECT_INHERIT_ACE
    SUCCESSFUL_ACCESS_ACE_FLAG = ldaptypes.ACE.SUCCESSFUL_ACCESS_ACE_FLAG

# ACE flags enum
# For an ACE, flags that indicate if the ObjectType and the InheritedObjecType are set with a GUID
# Since these two flags are the same for Allowed and Denied access, the same class will be used from 'ldaptypes'
# https://docs.microsoft.com/en-us/windows/win32/api/winnt/ns-winnt-access_allowed_object_ace
class OBJECT_ACE_FLAGS(Enum):
    ACE_OBJECT_TYPE_PRESENT = ldaptypes.ACCESS_ALLOWED_OBJECT_ACE.ACE_OBJECT_TYPE_PRESENT
    ACE_INHERITED_OBJECT_TYPE_PRESENT = ldaptypes.ACCESS_ALLOWED_OBJECT_ACE.ACE_INHERITED_OBJECT_TYPE_PRESENT


# Access Mask enum
# Access mask permits to encode principal's rights to an object. This is the rights the principal behind the specified SID has
# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-dtyp/7a53f60e-e730-4dfe-bbe9-b21b62eb790b
# https://docs.microsoft.com/en-us/windows/win32/api/iads/ne-iads-ads_rights_enum?redirectedfrom=MSDN
class ACCESS_MASK(Enum):
    # Generic Rights
    GenericRead = 0x80000000 # ADS_RIGHT_GENERIC_READ
    GenericWrite = 0x40000000 # ADS_RIGHT_GENERIC_WRITE
    GenericExecute = 0x20000000 # ADS_RIGHT_GENERIC_EXECUTE
    GenericAll = 0x10000000 # ADS_RIGHT_GENERIC_ALL

    # Maximum Allowed access type
    MaximumAllowed = 0x02000000

    # Access System Acl access type
    AccessSystemSecurity = 0x01000000 # ADS_RIGHT_ACCESS_SYSTEM_SECURITY

    # Standard access types
    Synchronize = 0x00100000 # ADS_RIGHT_SYNCHRONIZE
    WriteOwner = 0x00080000 # ADS_RIGHT_WRITE_OWNER
    WriteDACL = 0x00040000 # ADS_RIGHT_WRITE_DAC
    ReadControl = 0x00020000 # ADS_RIGHT_READ_CONTROL
    Delete = 0x00010000 # ADS_RIGHT_DELETE

    # Specific rights
    AllExtendedRights = 0x00000100 # ADS_RIGHT_DS_CONTROL_ACCESS
    ListObject = 0x00000080 # ADS_RIGHT_DS_LIST_OBJECT
    DeleteTree = 0x00000040 # ADS_RIGHT_DS_DELETE_TREE
    WriteProperties = 0x00000020 # ADS_RIGHT_DS_WRITE_PROP
    ReadProperties = 0x00000010 # ADS_RIGHT_DS_READ_PROP
    Self = 0x00000008 # ADS_RIGHT_DS_SELF
    ListChildObjects = 0x00000004 # ADS_RIGHT_ACTRL_DS_LIST
    DeleteChild = 0x00000002 # ADS_RIGHT_DS_DELETE_CHILD
    CreateChild = 0x00000001 # ADS_RIGHT_DS_CREATE_CHILD


# Simple permissions enum
# Simple permissions are combinaisons of extended permissions
# https://docs.microsoft.com/en-us/previous-versions/windows/it-pro/windows-server-2003/cc783530(v=ws.10)?redirectedfrom=MSDN
class SIMPLE_PERMISSIONS(Enum):
    FullControl = 0xf01ff
    Modify = 0x0301bf
    ReadAndExecute = 0x0200a9
    ReadAndWrite = 0x02019f
    Read = 0x20094
    Write = 0x200bc


# Mask ObjectType field enum
# Possible values for the Mask field in object-specific ACE (permitting to specify extended rights in the ObjectType field for example)
# Since these flags are the same for Allowed and Denied access, the same class will be used from 'ldaptypes'
# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-dtyp/c79a383c-2b3f-4655-abe7-dcbb7ce0cfbe
class ALLOWED_OBJECT_ACE_MASK_FLAGS(Enum):
    ControlAccess = ldaptypes.ACCESS_ALLOWED_OBJECT_ACE.ADS_RIGHT_DS_CONTROL_ACCESS
    CreateChild = ldaptypes.ACCESS_ALLOWED_OBJECT_ACE.ADS_RIGHT_DS_CREATE_CHILD
    DeleteChild = ldaptypes.ACCESS_ALLOWED_OBJECT_ACE.ADS_RIGHT_DS_DELETE_CHILD
    ReadProperty = ldaptypes.ACCESS_ALLOWED_OBJECT_ACE.ADS_RIGHT_DS_READ_PROP
    WriteProperty = ldaptypes.ACCESS_ALLOWED_OBJECT_ACE.ADS_RIGHT_DS_WRITE_PROP
    Self = ldaptypes.ACCESS_ALLOWED_OBJECT_ACE.ADS_RIGHT_DS_SELF

class DACLedit(object):
    """docstring for setrbcd"""

    def __init__(self, ldap_server, ldap_session, base_dn, target_sAMAccountName, target_SID, target_DN, target_sd, principal_sAMAccountName, principal_SID, principal_DN, ace_type, rights, rights_guid, inheritance):
        super(DACLedit, self).__init__()
        self.ldap_server = ldap_server
        self.ldap_session = ldap_session
        self.base_dn = base_dn

        self.target_sAMAccountName = target_sAMAccountName 
        self.target_SID = target_SID
        self.target_DN = target_DN
        self.principal_raw_security_descriptor = target_sd
        self.principal_security_descriptor = ldaptypes.SR_SECURITY_DESCRIPTOR(data=self.principal_raw_security_descriptor)


        self.principal_sAMAccountName = principal_sAMAccountName
        self.principal_SID = principal_SID
        self.principal_DN = principal_DN

        self.ace_type = ace_type
        self.rights = rights
        self.rights_guid = rights_guid
        self.inheritance = inheritance
        if self.inheritance:
            logging.warning("Objects with adminCount=1 will not inherit ACEs from their parent container/OU")

    # Main read funtion
    # Prints the parsed DACL
    def read(self):
        parsed_dacl = self.parseDACL(self.principal_security_descriptor['Dacl'])
        self.printparsedDACL(parsed_dacl)

    # Main write function
    # Attempts to add a new ACE to a DACL
    def write(self):
        # Creates ACEs with the specified GUIDs and the SID, or FullControl if no GUID is specified
        # Append the ACEs in the DACL locally
        if self.rights == "fullcontrol" and self.rights_guid is None:
            logging.warning(f"Adding FullControl to %s" % (self.target_SID if self.target_SID else self.target_DN))
            self.principal_security_descriptor['Dacl'].aces.append(self.create_ace(SIMPLE_PERMISSIONS.FullControl.value, self.principal_SID, self.ace_type))
        elif self.rights == "immutable" and self.rights_guid is None:
            logging.debug(f"Adding Delete and DeleteTree to %s" % (self.target_SID if self.target_SID else self.target_DN))
            self.principal_security_descriptor['Dacl'].aces.insert(0,self.create_ace(ACCESS_MASK.Delete.value + ACCESS_MASK.DeleteTree.value, self.principal_SID, ace_type="denied"))
        elif self.rights == "deletechild" and self.rights_guid is None:
            logging.debug(f"Adding DeleteChild to %s" % (self.target_SID if self.target_SID else self.target_DN))
            self.principal_security_descriptor['Dacl'].aces.insert(0,self.create_ace(ACCESS_MASK.DeleteChild.value, self.principal_SID, ace_type="denied"))
        else:
            for rights_guid in self.build_guids_for_rights():
                logging.debug("Adding %s (%s) to %s)" % (self.principal_SID, rights_guid, format_sid(self.target_SID)))
                self.principal_security_descriptor['Dacl'].aces.append(self.create_object_ace(rights_guid, self.principal_SID, self.ace_type))
        # Backups current DACL before add the new one
        # Effectively push the DACL with the new ACE
        return self.modify_secDesc_for_dn(self.target_DN, self.principal_security_descriptor)


    # Attempts to remove an ACE from the DACL
    # To do it, a new DACL is built locally with all the ACEs that must NOT BE removed, and this new DACL is pushed on the server
    def remove(self):
        compare_aces = []
        # Creates ACEs with the specified GUIDs and the SID, or FullControl if no GUID is specified
        # These ACEs will be used as comparison templates
        if self.rights == "fullcontrol" and self.rights_guid is None:
            compare_aces.append(self.create_ace(SIMPLE_PERMISSIONS.FullControl.value, self.principal_SID, self.ace_type))
        elif self.rights == "immutable" and self.rights_guid is None:
            logging.debug(f"Removing Delete and DeleteTree to %s" % (self.target_SID if self.target_SID else self.target_DN))
            compare_aces.append(self.create_ace(ACCESS_MASK.Delete.value + ACCESS_MASK.DeleteTree.value, self.principal_SID, ace_type="denied"))
        elif self.rights == "deletechild" and self.rights_guid is None:
            logging.debug(f"Removing DeleteChild to %s" % (self.target_SID if self.target_SID else self.target_DN))
            compare_aces.append(self.create_ace(ACCESS_MASK.DeleteChild.value, self.principal_SID, ace_type="denied"))
        else:
            for rights_guid in self.build_guids_for_rights():
                compare_aces.append(self.create_object_ace(rights_guid, self.principal_SID, self.ace_type))
        new_dacl = []
        i = 0
        dacl_must_be_replaced = False
        for ace in self.principal_security_descriptor['Dacl'].aces:
            ace_must_be_removed = False
            for compare_ace in compare_aces:
                # To be sure the good ACEs are removed, multiple fields are compared between the templates and the ACEs in the DACL
                #   - ACE type
                #   - ACE flags
                #   - Access masks
                #   - Revision
                #   - SubAuthorityCount
                #   - SubAuthority
                #   - IdentifierAuthority value
                if ace['AceType'] == compare_ace['AceType'] \
                    and ace['AceFlags'] == compare_ace['AceFlags']\
                    and ace['Ace']['Mask']['Mask'] == compare_ace['Ace']['Mask']['Mask']\
                    and ace['Ace']['Sid']['Revision'] == compare_ace['Ace']['Sid']['Revision']\
                    and ace['Ace']['Sid']['SubAuthorityCount'] == compare_ace['Ace']['Sid']['SubAuthorityCount']\
                    and ace['Ace']['Sid']['SubAuthority'] == compare_ace['Ace']['Sid']['SubAuthority']\
                    and ace['Ace']['Sid']['IdentifierAuthority']['Value'] == compare_ace['Ace']['Sid']['IdentifierAuthority']['Value']:
                    # If the ACE has an ObjectType, the GUIDs must match
                    if 'ObjectType' in ace['Ace'].fields.keys() and 'ObjectType' in compare_ace['Ace'].fields.keys():
                        if ace['Ace']['ObjectType'] == compare_ace['Ace']['ObjectType']:
                            ace_must_be_removed = True
                            dacl_must_be_replaced = True
                    else:
                        ace_must_be_removed = True
                        dacl_must_be_replaced = True
            # If the ACE doesn't match any ACEs from the template list, it is added to the DACL that will be pushed
            if not ace_must_be_removed:
                new_dacl.append(ace)
            elif logging.getLogger().level == logging.DEBUG:
                logging.debug("This ACE will be removed")
                self.printparsedACE(self.parseACE(ace))
            i += 1
        # If at least one ACE must been removed
        if dacl_must_be_replaced:
            self.principal_security_descriptor['Dacl'].aces = new_dacl
            self.modify_secDesc_for_dn(self.target_DN, self.principal_security_descriptor)
        else:
            logging.info("Nothing to remove...")

    
    # Attempts to retieve the SID and Distinguisehd Name from the sAMAccountName
    # Not used for the moment
    #   - samname : a sAMAccountName
    def get_user_info(self, samname):
        self.ldap_session.search(self.base_dn, '(sAMAccountName=%s)' % escape_filter_chars(samname), attributes=['objectSid'])
        try:
            dn = self.ldap_session.entries[0].entry_dn
            sid = format_sid(self.ldap_session.entries[0]['objectSid'].raw_values[0])
            return dn, sid
        except IndexError:
            logging.error('User not found in LDAP: %s' % samname)
            return False

    
    # Attempts to resolve a SID and return the corresponding samaccountname
    #   - sid : the SID to resolve
    def resolveSID(self, sid):
        # Tries to resolve the SID from the well known SIDs
        if sid in WELL_KNOWN_SIDS.keys():
            return WELL_KNOWN_SIDS[sid]
        # Tries to resolve the SID from the LDAP domain dump
        else:
            self.ldap_session.search(self.base_dn, '(objectSid=%s)' % sid, attributes=['samaccountname'])
            try:
                dn = self.ldap_session.entries[0].entry_dn
                samname = self.ldap_session.entries[0]['samaccountname']
                return samname
            except IndexError:
                logging.debug('SID not found in LDAP: %s' % sid)
                return ""

    
    # Parses a full DACL
    #   - dacl : the DACL to parse, submitted in a Security Desciptor format
    def parseDACL(self, dacl):
        parsed_dacl = []
        logging.info("Parsing DACL")
        i = 0
        for ace in dacl['Data']:
            parsed_ace = self.parseACE(ace)
            parsed_dacl.append(parsed_ace)
            i += 1
        return parsed_dacl

    
    # Parses an access mask to extract the different values from a simple permission
    # https://stackoverflow.com/questions/28029872/retrieving-security-descriptor-and-getting-number-for-filesystemrights
    #   - fsr : the access mask to parse
    def parsePerms(self, fsr):
        _perms = []
        for PERM in SIMPLE_PERMISSIONS:
            if (fsr & PERM.value) == PERM.value:
                _perms.append(PERM.name)
                fsr = fsr & (not PERM.value)
        for PERM in ACCESS_MASK:
            if fsr & PERM.value:
                _perms.append(PERM.name)
        return _perms

    
    # Parses a specified ACE and extract the different values (Flags, Access Mask, Trustee, ObjectType, InheritedObjectType)
    #   - ace : the ACE to parse
    def parseACE(self, ace):
        # For the moment, only the Allowed and Denied Access ACE are supported
        if ace['TypeName'] in [ "ACCESS_ALLOWED_ACE", "ACCESS_ALLOWED_OBJECT_ACE", "ACCESS_DENIED_ACE", "ACCESS_DENIED_OBJECT_ACE" ]:
            parsed_ace = {}
            parsed_ace['ACE Type'] = ace['TypeName']
            # Retrieves ACE's flags
            _ace_flags = []
            for FLAG in ACE_FLAGS:
                if ace.hasFlag(FLAG.value):
                    _ace_flags.append(FLAG.name)
            parsed_ace['ACE flags'] = ", ".join(_ace_flags) or "None"

            # For standard ACE
            # Extracts the access mask (by parsing the simple permissions) and the principal's SID
            if ace['TypeName'] in [ "ACCESS_ALLOWED_ACE", "ACCESS_DENIED_ACE" ]:
                parsed_ace['Access mask'] = "%s (0x%x)" % (", ".join(self.parsePerms(ace['Ace']['Mask']['Mask'])), ace['Ace']['Mask']['Mask'])
                parsed_ace['Trustee (SID)'] = "%s (%s)" % (self.resolveSID(ace['Ace']['Sid'].formatCanonical()) or "UNKNOWN", ace['Ace']['Sid'].formatCanonical())

            # For object-specific ACE
            elif ace['TypeName'] in [ "ACCESS_ALLOWED_OBJECT_ACE", "ACCESS_DENIED_OBJECT_ACE" ]:
                # Extracts the mask values. These values will indicate the ObjectType purpose
                _access_mask_flags = []
                for FLAG in ALLOWED_OBJECT_ACE_MASK_FLAGS:
                    if ace['Ace']['Mask'].hasPriv(FLAG.value):
                        _access_mask_flags.append(FLAG.name)
                parsed_ace['Access mask'] = ", ".join(_access_mask_flags)
                # Extracts the ACE flag values and the trusted SID
                _object_flags = []
                for FLAG in OBJECT_ACE_FLAGS:
                    if ace['Ace'].hasFlag(FLAG.value):
                        _object_flags.append(FLAG.name)
                parsed_ace['Flags'] = ", ".join(_object_flags) or "None"
                # Extracts the ObjectType GUID values
                if ace['Ace']['ObjectTypeLen'] != 0:
                    obj_type = bin_to_string(ace['Ace']['ObjectType']).lower()
                    try:
                        parsed_ace['Object type (GUID)'] = "%s (%s)" % (OBJECT_TYPES_GUID[obj_type], obj_type)
                    except KeyError:
                        parsed_ace['Object type (GUID)'] = "UNKNOWN (%s)" % obj_type
                # Extracts the InheritedObjectType GUID values
                if ace['Ace']['InheritedObjectTypeLen'] != 0:
                    inh_obj_type = bin_to_string(ace['Ace']['InheritedObjectType']).lower()
                    try:
                        parsed_ace['Inherited type (GUID)'] = "%s (%s)" % (OBJECT_TYPES_GUID[inh_obj_type], inh_obj_type)
                    except KeyError:
                        parsed_ace['Inherited type (GUID)'] = "UNKNOWN (%s)" % inh_obj_type
                # Extract the Trustee SID (the object that has the right over the DACL bearer)
                parsed_ace['Trustee (SID)'] = "%s (%s)" % (self.resolveSID(ace['Ace']['Sid'].formatCanonical()) or "UNKNOWN", ace['Ace']['Sid'].formatCanonical())

        else:
            # If the ACE is not an access allowed
            logging.debug("ACE Type (%s) unsupported for parsing yet, feel free to contribute" % ace['TypeName'])
            parsed_ace = {}
            parsed_ace['ACE type'] = ace['TypeName']
            _ace_flags = []
            for FLAG in ACE_FLAGS:
                if ace.hasFlag(FLAG.value):
                    _ace_flags.append(FLAG.name)
            parsed_ace['ACE flags'] = ", ".join(_ace_flags) or "None"
            parsed_ace['DEBUG'] = "ACE type not supported for parsing by dacleditor.py, feel free to contribute"
        return parsed_ace


    # Prints a full DACL by printing each parsed ACE
    #   - parsed_dacl : a parsed DACL from parseDACL()
    def printparsedDACL(self, parsed_dacl):
        # Attempts to retrieve the principal's SID if it's a write action
        if self.principal_SID is None and self.principal_sAMAccountName or self.principal_DN:
            if self.principal_sAMAccountName is not None:
                _lookedup_principal = self.principal_sAMAccountName
                self.ldap_session.search(self.base_dn, '(sAMAccountName=%s)' % escape_filter_chars(_lookedup_principal), attributes=['objectSid'])
            elif self.principal_DN is not None:
                _lookedup_principal = self.principal_DN
                self.ldap_session.search(self.base_dn, '(distinguishedName=%s)' % _lookedup_principal, attributes=['objectSid'])
            try:
                self.principal_SID = format_sid(self.ldap_session.entries[0]['objectSid'].raw_values[0])
            except IndexError:
                logging.error('Principal not found in LDAP (%s)' % _lookedup_principal)
                return False
            logging.debug("Found principal SID to write in ACE(s): %s" % self.principal_SID)

        logging.info("Printing parsed DACL")
        i = 0
        # If a principal has been specified, only the ACE where he is the trustee will be printed
        if self.principal_SID is not None:
            logging.info("Filtering results for SID (%s)" % self.principal_SID)
        for parsed_ace in parsed_dacl:
            print_ace = True
            if self.principal_SID is not None:
                try:
                    if self.principal_SID not in parsed_ace['Trustee (SID)']:
                        print_ace = False
                except Exception as e:
                    logging.error("Error filtering ACE, probably because of ACE type unsupported for parsing yet (%s)" % e)
            if print_ace:
                logging.info("  %-28s" % "ACE[%d] info" % i)
                self.printparsedACE(parsed_ace)
            i += 1


    # Prints properly a parsed ACE
    #   - parsed_ace : a parsed ACE from parseACE()
    def printparsedACE(self, parsed_ace):
        elements_name = list(parsed_ace.keys())
        for attribute in elements_name:
            logging.info("    %-26s: %s" % (attribute, parsed_ace[attribute]))


    # Retrieves the GUIDs for the specified rights
    def build_guids_for_rights(self):
        _rights_guids = []
        if self.rights_guid is not None:
            _rights_guids = [self.rights_guid]
        elif self.rights == "writemembers":
            _rights_guids = [RIGHTS_GUID.WriteMembers.value]
        elif self.rights == "resetpassword":
            _rights_guids = [RIGHTS_GUID.ResetPassword.value]
        elif self.rights == "dcsync":
            _rights_guids = [RIGHTS_GUID.DS_Replication_Get_Changes.value, RIGHTS_GUID.DS_Replication_Get_Changes_All.value]
        logging.debug('Built GUID: %s', _rights_guids)
        return _rights_guids


    # Attempts to push the locally built DACL to the remote server into the security descriptor of the specified principal
    # The target principal is specified with its Distinguished Name
    #   - dn : the principal's Distinguished Name to modify
    #   - secDesc : the Security Descriptor with the new DACL to push
    def modify_secDesc_for_dn(self, dn, secDesc):
        data = secDesc.getData()
        controls = security_descriptor_control(sdflags=0x04)
        logging.debug('Attempts to modify the Security Descriptor.')
        self.ldap_session.modify(dn, {'nTSecurityDescriptor': (ldap3.MODIFY_REPLACE, [data])}, controls=controls)
        if self.ldap_session.result['result'] == 0:
            logging.debug('DACL modified successfully!')
            return True
        else:
            if self.ldap_session.result['result'] == 50:
                logging.error('Could not modify object, the server reports insufficient rights: %s',
                              self.ldap_session.result['message'])
            elif self.ldap_session.result['result'] == 19:
                logging.error('Could not modify object, the server reports a constrained violation: %s',
                              self.ldap_session.result['message'])
            else:
                logging.error('The server returned an error: %s', self.ldap_session.result['message'])
            return False


    # Builds a standard ACE for a specified access mask (rights) and a specified SID (the principal who obtains the right)
    # https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-dtyp/72e7c7ea-bc02-4c74-a619-818a16bf6adb
    #   - access_mask : the allowed access mask
    #   - sid : the principal's SID
    #   - ace_type : the ACE type (allowed or denied)
    def create_ace(self, access_mask, sid, ace_type):
        nace = ldaptypes.ACE()
        if ace_type == "allowed":
            nace['AceType'] = ldaptypes.ACCESS_ALLOWED_ACE.ACE_TYPE
            acedata = ldaptypes.ACCESS_ALLOWED_ACE()
        else:
            nace['AceType'] = ldaptypes.ACCESS_DENIED_ACE.ACE_TYPE
            acedata = ldaptypes.ACCESS_DENIED_ACE()
        if self.inheritance:
            nace['AceFlags'] = ldaptypes.ACE.OBJECT_INHERIT_ACE + ldaptypes.ACE.CONTAINER_INHERIT_ACE
        else:
            nace['AceFlags'] = 0x00
        acedata['Mask'] = ldaptypes.ACCESS_MASK()
        acedata['Mask']['Mask'] = access_mask
        acedata['Sid'] = ldaptypes.LDAP_SID()
        acedata['Sid'].fromCanonical(sid)
        nace['Ace'] = acedata
        logging.debug('ACE created.')
        return nace


    # Builds an object-specific for a specified ObjectType (an extended right, a property, etc, to add) for a specified SID (the principal who obtains the right)
    # The Mask is "ADS_RIGHT_DS_CONTROL_ACCESS" (the ObjectType GUID will identify an extended access right)
    # https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-dtyp/c79a383c-2b3f-4655-abe7-dcbb7ce0cfbe
    #   - privguid : the ObjectType (an Extended Right here)
    #   - sid : the principal's SID
    #   - ace_type : the ACE type (allowed or denied)
    def create_object_ace(self, privguid, sid, ace_type):
        nace = ldaptypes.ACE()
        if ace_type == "allowed":
            nace['AceType'] = ldaptypes.ACCESS_ALLOWED_OBJECT_ACE.ACE_TYPE
            acedata = ldaptypes.ACCESS_ALLOWED_OBJECT_ACE()
        else:
            nace['AceType'] = ldaptypes.ACCESS_DENIED_OBJECT_ACE.ACE_TYPE
            acedata = ldaptypes.ACCESS_DENIED_OBJECT_ACE()
        if self.inheritance:
            nace['AceFlags'] = ldaptypes.ACE.OBJECT_INHERIT_ACE + ldaptypes.ACE.CONTAINER_INHERIT_ACE
        else:
            nace['AceFlags'] = 0x00
        acedata['Mask'] = ldaptypes.ACCESS_MASK()
        # WriteMembers not an extended right, we need read and write mask on the attribute (https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-dtyp/c79a383c-2b3f-4655-abe7-dcbb7ce0cfbe)
        if privguid == RIGHTS_GUID.WriteMembers.value:
            acedata['Mask']['Mask'] = ldaptypes.ACCESS_ALLOWED_OBJECT_ACE.ADS_RIGHT_DS_READ_PROP + ldaptypes.ACCESS_ALLOWED_OBJECT_ACE.ADS_RIGHT_DS_WRITE_PROP
        # Other rights in this script are extended rights and need the DS_CONTROL_ACCESS mask
        else:
            acedata['Mask']['Mask'] = ldaptypes.ACCESS_ALLOWED_OBJECT_ACE.ADS_RIGHT_DS_CONTROL_ACCESS
        acedata['ObjectType'] = string_to_bin(privguid)
        acedata['InheritedObjectType'] = b''
        acedata['Sid'] = ldaptypes.LDAP_SID()
        acedata['Sid'].fromCanonical(sid)
        assert sid == acedata['Sid'].formatCanonical()
        # This ACE flag verifes if the ObjectType is valid
        acedata['Flags'] = ldaptypes.ACCESS_ALLOWED_OBJECT_ACE.ACE_OBJECT_TYPE_PRESENT
        nace['Ace'] = acedata
        logging.debug('Object-specific ACE created.')
        return nace



#!/usr/bin/env python3

EDRS = [
	{
	  "name": "Bitdefender",
	  "services": [
	    {
	      "name": "bdredline_agent",
	      "description": "Bitdefender Agent RedLine Service"
	    },
	    {
	      "name": "BDAuxSrv",
	      "description": "Bitdefender Auxiliary Service"
	    },
	    {
	      "name": "UPDATESRV",
	      "description": "Bitdefender Desktop Update Service"
	    },
	    {
	      "name": "VSSERV",
	      "description": "Bitdefender Virus Shield"
	    },
	    {
	      "name": "bdredline",
	      "description": "Bitdefender RedLine Service"
	    }
	  ],
	  "pipes": [
	    {
	      "name": "local\\msgbus\\antitracker.low\\*",
	      "processes": [
	        "bdagent.exe"
	      ]
	    },
	    {
	      "name": "local\\msgbus\\aspam.actions.low\\*",
	      "processes": [
	        "bdagent.exe"
	      ]
	    },
	    {
	      "name": "local\\msgbus\\bd.process.broker.pipe",
	      "processes": [
	        "bdagent.exe",
	        "bdservicehost.exe",
	        "updatesrv.exe"
	      ]
	    },
	    {
	      "name": "local\\msgbus\\bdagent*",
	      "processes": [
	        "bdagent.exe"
	      ]
	    },
	    {
	      "name": "local\\msgbus\\bdauxsrv",
	      "processes": [
	        "bdagent.exe",
	        "bdntwrk.exe"
	      ]
	    }
	  ]
	},
	{
	  "name": "Windows Defender",
	  "services": [
	    {
	      "name": "WinDefend",
	      "description": "Windows Defender Antivirus Service"
	    },
	    {
	      "name": "Sense",
	      "description": "Windows Defender Advanced Threat Protection Service"
	    },
	    {
	      "name": "WdNisSvc",
	      "description": "Windows Defender Antivirus Network Inspection Service"
	    },
	    {
	    	"name": "WdNisDrv",
	    	"description": "Windows Defender Antivirus Network Inspection System Driver"
	    }
	  ],
	  "pipes": []
	},
	{
	  "name": "ESET",
	  "services": [
	    {
	      "name": "ekm",
	      "description": "ESET"
	    },
	    {
	      "name": "epfw",
	      "description": "ESET"
	    },
	    {
	      "name": "epfwlwf",
	      "description": "ESET"
	    },
	    {
	      "name": "epfwwfp",
	      "description": "ESET"
	    },
	    {
	      "name": "EraAgentSvc",
	      "description": "ESET"
	    }
	  ],
	  "pipes": [
	    {
	      "name": "nod_scriptmon_pipe",
	      "processes": [
	        ""
	      ]
	    }
	  ]
	},
	{
	  "name": "CrowdStrike",
	  "services": [
	    {
	      "name": "CSFalconService",
	      "description": "CrowdStrike Falcon Sensor Service"
	    }
	  ],
	  "pipes": [
	    {
	      "name": "CrowdStrike\\{*",
	      "processes": [
	        "CSFalconContainer.exe",
	        "CSFalconService.exe"
	      ]
	    }
	  ]
	},
	{
	  "name": "SentinelOne",
	  "services": [
	    {
	      "name": "SentinelAgent",
	      "description": "SentinelOne Endpoint Protection Agent"
	    },
	    {
	      "name": "SentinelStaticEngine",
	      "description": "Manage static engines for SentinelOne Endpoint Protection"
	    },
	    {
	      "name": "LogProcessorService",
	      "description": "Manage logs for SentinelOne Endpoint Protection"
	    },
	    {
	    	"name": "SentinelHelperService",
	    	"description": "SentinelOne Helper Service"
	    }
	  ],
	  "pipes": [
	    {
	      "name": "SentinelAgentWorkerCert.*",
	      "processes": [
	        ""
	      ]
	    },
	    {
	      "name": "DFIScanner.Etw.*",
	      "processes": [
	        "SentinelStaticEngine.exe"
	      ]
	    },
	    {
	      "name": "DFIScanner.Inline.*",
	      "processes": [
	        "SentinelAgent.exe"
	      ]
	    }
	  ]
	},
	{
	  "name": "Carbon Black App Control",
	  "services": [
	    {
	      "name": "Parity",
	      "description": "Carbon Black App Control Agent"
	    }
	  ],
	  "pipes": []
	},
	{
	  "name": "Cybereason",
	  "services": [
	    {
	      "name": "CybereasonActiveProbe",
	      "description": "Cybereason Active Probe"
	    },
	    {
	      "name": "CybereasonCRS",
	      "description": "Cybereason Anti-Ransomware"
	    },
	    {
	      "name": "CybereasonBlocki",
	      "description": "Cybereason Execution Prevention"
	    }
	  ],
	  "pipes": [
	    {
	      "name": "CybereasonAPConsoleMinionHostIpc_*",
	      "processes": [
	        "minionhost.exe"
	      ]
	    },
	    {
	      "name": "CybereasonAPServerProxyIpc_*",
	      "processes": [
	        "minionhost.exe"
	      ]
	    }
	  ]
	},
	{
	  "name": "Symantec Endpoint Protection",
	  "services": [
	    {
	      "name": "SepMasterService",
	      "description": "Symantec Endpoint Protection"
	    },
	    {
	      "name": "SepScanService",
	      "description": "Symantec Endpoint Protection Scan Services"
	    },
	    {
	      "name": "SNAC",
	      "description": "Symantec Network Access Control"
	    }
	  ],
	  "pipes": []
	},
	{
	  "name": "Carbon Black",
	  "services": [
	    {
	      "name": "CbDefense",
	      "description": "Carbon Black Cloud Sensor service"
	    },
	    {
	      "name": "CbDefenseWSC",
	      "description": "Carbon Black Cloud Sensor WSC service"
	    }
	  ],
	  "pipes": []
	},
	{
	  "name": "SysMon",
	  "services": [
	    {
	      "name": "Sysmon64",
	      "description": "System Monitor service"
	    },
		{
	      "name": "Sysmon",
	      "description": "System Monitor service"
	    }
	  ],
	  "pipes": []
	},
	{
		"name": "Trend Micro",
		"services": [
			{
				"name": "ds_agent",
				"description": "Trend Micro Deep Security Agent"
			}
		],
		"pipes": []
	},
	{
	  "name": "Sophos Intercept X",
	  "services": [
	    {
	      "name": "SntpService",
	      "description": "Sophos Network Threat Protection"
	    },
	    {
	      "name": "Sophos Endpoint Defense Service",
	      "description": "Sophos Endpoint Defense Service"
	    },
	    {
	      "name": "Sophos File Scanner Service",
	      "description": "Sophos File Scanner Service"
	    },
	    {
	      "name": "Sophos Health Service",
	      "description": "Sophos Health Service"
	    },
	    {
	      "name": "Sophos Live Query",
	      "description": "Sophos Live Query"
	    },
	    {
	      "name": "Sophos Managed Threat Response",
	      "description": "Sophos Managed Threat Response"
	    },
	    {
	      "name": "Sophos MCS Agent",
	      "description": "Sophos MCS Agent"
	    },
	    {
	      "name": "Sophos MCS Client",
	      "description": "Sophos MCS Client"
	    },
	    {
	      "name": "Sophos System Protection Service",
	      "description": "Sophos System Protection Service"
	    }
	  ],
	  "pipes": []
	},
	{
		"name": "Cylance",
		"services": [
			{
				"name": "CylanceDrv",
				"description": "CylanceDrv"
			},
			{
				"name": "CylanceSvc",
				"description": "Cylance Unified Agent"
			},
			{
				"name": "CyOptics",
				"description": "Cylance Optics"
			},
			{
				"name": "CyOpticsDrv",
				"description": "CyOpticsDrv"
			},
			{
				"name": "CyProtectDrv",
				"description": "CyProtectDrv"
			}
		],
		"pipes": []
	},
	{
		"name": "Trend Micro",
		"services": [
			{
				"name": "TmPreFilter",
				"description": "Trend Micro PreFilter"
			},
			{
				"name": "tmumh",
				"description": "Trend Micro User Mode Hook Driver"
			},
			{
				"name": "tmusa",
				"description": "Trend Micro Osprey Driver"
			},
			{
				"name": "tmWfp",
				"description": "Trend Micro WFP Callout Driver"
			},
			{
				"name": "TmWSCSvc",
				"description": "Apex One NT WSC Service"
			},
			{
				"name": "Trend Micro Endpoint Basecamp",
				"description": "Trend Micro Endpoint Basecamp"
			},
			{
				"name": "Trend Micro Web Service Communicator",
				"description": "Trend Micro Web Service Communicator"
			}
		],
		"pipes": []
	}
]

class EDR:
	def __init__(self):
		self.names = []
		self.services = []

		for edr in EDRS:
			self.names.append(edr["name"])
			for svc in edr["services"]:
				self.services.append(svc["name"])

	def service_exist(self, product) -> bool:
		if product in self.services:
			return True
		else:
			return False


#!/usr/bin/env python3

import json
from collections import abc
from datetime import timedelta

def make_serializable(data):
    if isinstance(data, abc.Mapping):
        return {key: make_serializable(value) for key, value in data.items()}
    elif isinstance(data, list):
        return [make_serializable(element) for element in data]
    elif isinstance(data, bytes):
        return data.decode('utf-8', errors='ignore')
    elif isinstance(data, timedelta):
        return str(data)
    else:
        return data


#!/usr/bin/env python3
import logging
import re
import ldap3
import datetime
from impacket.ldap import ldaptypes
from ldap3.protocol.formatters.formatters import format_sid
from ldap3.protocol.microsoft import security_descriptor_control

from powerview.utils.helpers import (
    strip_entry,
    is_admin_sid,
    format_datetime,
)
from powerview.utils.constants import (
    WELL_KNOWN_SIDS,
    ACTIVE_DIRECTORY_RIGHTS,
    EXTENDED_RIGHTS_NAME_MAP
)

# Exchange-specific constants
EXCHANGE_OBJECT_TYPES = {
    "msExchExchangeServer": "Exchange Server",
    "msExchMailboxDatabase": "Mailbox Database",
    "msExchMailboxServer": "Mailbox Server",
    "msExchActiveSync": "ActiveSync",
    "msExchMailboxUser": "Mailbox User",
    "msExchPublicFolder": "Public Folder"
}

EXCHANGE_ROLES = {
    "Organization Management": "Most powerful Exchange role, equivalent to Domain Admin in Exchange context",
    "Recipient Management": "Can create and modify Exchange recipients",
    "Exchange Trusted Subsystem": "Has extensive Exchange permissions",
    "Public Folder Management": "Can manage public folders",
    "Help Desk": "Can reset passwords and manage mail recipient properties",
    "Discovery Management": "Can perform mailbox searches and exports",
    "Mailbox Import Export": "Can import and export mailbox content"
}

EXCHANGE_EXTENDED_RIGHTS = {
    "1131f6aa-9c07-11d1-f79f-00c04fc2dcd2": "Send-As Right",
    "1131f6ad-9c07-11d1-f79f-00c04fc2dcd2": "Receive-As Right",
    "ab721a54-1e2f-11d0-9819-00aa0040529b": "Open-Mailbox Right",
    "dc50a4f8-6fc8-4b5f-942f-2b103221a85d": "Send-To Right"
}

# Combining Exchange-specific extended rights with common AD extended rights
ALL_EXTENDED_RIGHTS = {**EXTENDED_RIGHTS_NAME_MAP, **EXCHANGE_EXTENDED_RIGHTS}

class ExchangeSecurity:
    """
    Class for parsing and analyzing security descriptors on Exchange objects
    """
    def __init__(self, security_descriptor):
        self.sd = ldaptypes.SR_SECURITY_DESCRIPTOR()
        self.sd.fromString(security_descriptor)
        
        self.owner = format_sid(self.sd["OwnerSid"].getData())
        self.owner_name = None
        self.aces = {}
        
        # Process all ACEs in the security descriptor
        if self.sd["Dacl"]:
            aces = self.sd["Dacl"]["Data"]
            for ace in aces:
                sid = format_sid(ace["Ace"]["Sid"].getData())
                
                if sid not in self.aces:
                    self.aces[sid] = {
                        "rights": ACTIVE_DIRECTORY_RIGHTS(0),
                        "extended_rights": [],
                        "inherited": bool(ace["AceFlags"] & 0x10),  # INHERITED_ACE flag
                        "object_type": None
                    }
                
                # Process allowed access ACEs
                if ace["AceType"] == ldaptypes.ACCESS_ALLOWED_ACE.ACE_TYPE:
                    self.aces[sid]["rights"] |= ACTIVE_DIRECTORY_RIGHTS(ace["Ace"]["Mask"]["Mask"])
                
                # Process object-specific ACEs (like extended rights)
                elif ace["AceType"] == ldaptypes.ACCESS_ALLOWED_OBJECT_ACE.ACE_TYPE:
                    if ace["Ace"]["Flags"] & 0x01:  # OBJECT_TYPE_PRESENT flag
                        uuid = str(ace["Ace"]["ObjectType"]).lower()
                        self.aces[sid]["extended_rights"].append(uuid)
                        self.aces[sid]["object_type"] = uuid
    
    def get_owner_name(self, resolver_func=None):
        """Get the name of the security descriptor owner"""
        if not self.owner_name and resolver_func:
            self.owner_name = resolver_func(self.owner)
        return self.owner_name or self.owner
    
    def get_ace_details(self, resolver_func=None):
        """Get comprehensive details about all ACEs"""
        result = []
        
        for sid, ace_data in self.aces.items():
            principal_name = resolver_func(sid) if resolver_func else sid
            
            # Create a human-readable description of rights
            rights_description = []
            for right_name, right_value in ACTIVE_DIRECTORY_RIGHTS.__members__.items():
                if ace_data["rights"] & right_value:
                    rights_description.append(right_name)
            
            # Create a human-readable description of extended rights
            extended_rights_desc = []
            for right_id in ace_data["extended_rights"]:
                right_name = ALL_EXTENDED_RIGHTS.get(right_id, f"Unknown Right ({right_id})")
                extended_rights_desc.append(right_name)
            
            result.append({
                "Principal": principal_name,
                "SID": sid,
                "Rights": ", ".join(rights_description) if rights_description else "None",
                "ExtendedRights": ", ".join(extended_rights_desc) if extended_rights_desc else "None",
                "Inherited": ace_data["inherited"]
            })
        
        return result
    
    def check_dangerous_permissions(self):
        """Check for permissions that could be abused in Exchange"""
        dangerous = []
        
        for sid, ace_data in self.aces.items():
            # Check for dangerous standard rights
            if ace_data["rights"] & ACTIVE_DIRECTORY_RIGHTS.WRITE_DAC:
                dangerous.append({"SID": sid, "Issue": "WriteDACL permission - Can modify permissions"})
            
            if ace_data["rights"] & ACTIVE_DIRECTORY_RIGHTS.WRITE_OWNER:
                dangerous.append({"SID": sid, "Issue": "WriteOwner permission - Can take ownership"})
            
            if ace_data["rights"] & ACTIVE_DIRECTORY_RIGHTS.GENERIC_ALL:
                dangerous.append({"SID": sid, "Issue": "GenericAll permission - Full control"})
            
            # Check for dangerous extended rights
            for right_id in ace_data["extended_rights"]:
                if right_id in EXCHANGE_EXTENDED_RIGHTS:
                    right_name = EXCHANGE_EXTENDED_RIGHTS[right_id]
                    dangerous.append({"SID": sid, "Issue": f"Exchange right: {right_name}"})
        
        return dangerous


class ExchangeEnum:
    """
    Class for enumerating and analyzing Exchange objects and permissions in Active Directory
    """
    def __init__(self, powerview):
        self.powerview = powerview
        self.ldap_session = self.powerview.conn.ldap_session
        self.ldap_server = self.powerview.conn.ldap_server
        self.root_dn = self.powerview.root_dn
        self.configuration_dn = self.powerview.configuration_dn
    
    def get_exchange_servers(self, properties=None, identity=None, searchbase=None, sd_flag=None, 
                          search_scope=ldap3.SUBTREE, no_cache=False, no_vuln_check=False, raw=False):
        """
        Find Exchange servers in the domain
        
        Args:
            properties: Properties to retrieve (optional)
            identity: Server name to target (optional)
            searchbase: Search base for the LDAP query (optional)
            search_scope: LDAP search scope (default: SUBTREE)
            no_cache: Whether to use cache (default: False)
            no_vuln_check: Whether to skip vulnerability checks (default: False)
            raw: Whether to return raw LDAP entries (default: False)
        Returns:
            List of Exchange servers with their properties
        """
        if not properties:
            properties = [
                "name",
                "cn",
                "objectClass",
                "objectCategory",
                "objectGUID",
                "distinguishedName",
                "whenCreated",
                "whenChanged",
                "serialNumber",
                "serverRole",
                "legacyExchangeDN",
                "networkAddress",
                "msExchServerRole",
                "msExchCurrentServerRoles",
                "msExchVersion",
                "msExchProductID",
                "msExchServerSite",
                "msExchServerFaultZone",
                "msExchInstallPath",
                "msExchDataPath",
                "msExchComponentStates",
                "msExchMinAdminVersion",
                "msExchMailboxRelease",
                "msExchServerInternalTLSCert",
                "nTSecurityDescriptor"
            ]
        
        if not searchbase:
            searchbase = self.configuration_dn

        if sd_flag:
            sd_flag = security_descriptor_control(sdflags=sd_flag)
        
        search_filter = "(&(objectClass=msExchExchangeServer)(adminDisplayName=*))"
        
        if identity:
            search_filter = f"(&{search_filter}(|(name={identity})(dNSHostName={identity})))"
        
        logging.debug(f"[Exchange] LDAP search base: {searchbase}")
        logging.debug(f"[Exchange] LDAP search filter: {search_filter}")
        
        return self.ldap_session.extend.standard.paged_search(
            searchbase, 
            search_filter, 
            attributes=properties, 
            paged_size=1000, 
            generator=True,
            controls=sd_flag,
            search_scope=search_scope,
            no_cache=no_cache,
            no_vuln_check=no_vuln_check,
            raw=raw
        )
    
    def get_exchange_mailboxes(self, properties=None, identity=None, searchbase=None, sd_flag=None, 
                            search_scope=ldap3.SUBTREE, no_cache=False, no_vuln_check=False, raw=False):
        """
        Find Exchange mailboxes in the domain
        
        Args:
            properties: Properties to retrieve (optional)
            identity: User name to target (optional)
            searchbase: Search base for the LDAP query (optional)
            search_scope: LDAP search scope (default: SUBTREE)
            no_cache: Whether to use cache (default: False)
            no_vuln_check: Whether to skip vulnerability checks (default: False)
            raw: Whether to return raw LDAP entries (default: False)
        Returns:
            List of Exchange mailboxes with their properties
        """
        if not properties:
            properties = [
                "name",
                "mail", 
                "proxyAddresses",
                "objectGUID", 
                "objectSid",
                "legacyExchangeDN",
                "homeMDB",
                "msExchHomeServerName",
                "msExchMailboxGuid",
                "msExchMailboxSecurityDescriptor",
                "msExchDelegateListLink",
                "nTSecurityDescriptor",
                "whenCreated",
                "whenChanged"
            ]
        
        if not searchbase:
            searchbase = self.configuration_dn

        if sd_flag:
            sd_flag = security_descriptor_control(sdflags=sd_flag)
        
        # Look for users with Exchange mailboxes
        search_filter = "(&(objectCategory=person)(objectClass=user)(msExchMailboxGuid=*))"
        
        if identity:
            search_filter = f"(&{search_filter}(|(name={identity})(sAMAccountName={identity})(mail={identity})))"
        
        logging.debug(f"[Exchange] LDAP search base: {searchbase}")
        logging.debug(f"[Exchange] LDAP search filter: {search_filter}")
        
        return self.ldap_session.extend.standard.paged_search(
            searchbase, 
            search_filter, 
            attributes=properties, 
            paged_size=1000, 
            generator=True,
            controls=sd_flag,
            search_scope=search_scope,
            no_cache=no_cache,
            no_vuln_check=no_vuln_check,
            raw=raw
        )
    
    def get_exchange_databases(self, properties=None, identity=None, searchbase=None, sd_flag=None, 
                            search_scope=ldap3.SUBTREE, no_cache=False, no_vuln_check=False, raw=False):
        """
        Find Exchange mailbox databases in the domain
        
        Args:
            properties: Properties to retrieve (optional)
            identity: Database name to target (optional)
            searchbase: Search base for the LDAP query (optional)
            search_scope: LDAP search scope (default: SUBTREE)
            no_cache: Whether to use cache (default: False)
            no_vuln_check: Whether to skip vulnerability checks (default: False)
            raw: Whether to return raw LDAP entries (default: False)
        Returns:
            List of Exchange mailbox databases with their properties
        """
        if not properties:
            properties = [
                "objectClass",
                "cn",
                "distinguishedName",
                "displayName",
                "adminDisplayName",
                "objectCategory",
                "name",
                "objectGUID",
                "homeMDBBL",
                "msExchOwningServer",
                "msExchVersion",
                "msExchDatabaseCreated",
                "whenCreated",
                "whenChanged"
            ]
        
        if not searchbase:
            searchbase = self.configuration_dn

        if sd_flag:
            sd_flag = security_descriptor_control(sdflags=sd_flag)
        
        search_filter = "(objectClass=msExchMDB)"
        
        if identity:
            search_filter = f"(&{search_filter}(|(name={identity})(msExchDatabaseName={identity})))"
        
        logging.debug(f"[Exchange] LDAP search base: {searchbase}")
        logging.debug(f"[Exchange] LDAP search filter: {search_filter}")
        
        return self.ldap_session.extend.standard.paged_search(
            searchbase, 
            search_filter, 
            attributes=properties, 
            paged_size=1000, 
            generator=True,
            controls=sd_flag,
            search_scope=search_scope,
            no_cache=no_cache,
            no_vuln_check=no_vuln_check,
            raw=raw
        )
    
    def get_exchange_organization(self, properties=None, searchbase=None, sd_flag=None,
                               search_scope=ldap3.SUBTREE, no_cache=False, no_vuln_check=False, raw=False):
        """
        Find Exchange organization configuration in the domain
        
        Args:
            properties: Properties to retrieve (optional)
            searchbase: Search base for the LDAP query (optional)
            search_scope: LDAP search scope (default: SUBTREE)
            no_cache: Whether to use cache (default: False)
            no_vuln_check: Whether to skip vulnerability checks (default: False)
            raw: Whether to return raw LDAP entries (default: False)
        Returns:
            Exchange organization configuration
        """
        if not properties:
            properties = [
                "name",
                "objectGUID",
                "msExchProductID", 
                "msExchVersion",
                "msExchOrganizationName",
                "whenCreated",
                "whenChanged"
            ]
        
        if not searchbase:
            # Exchange organization containers are in the Configuration naming context
            searchbase = f"CN=Microsoft Exchange,CN=Services,{self.configuration_dn}"

        if sd_flag:
            sd_flag = security_descriptor_control(sdflags=sd_flag)
        
        search_filter = "(objectClass=msExchOrganizationContainer)"
        
        logging.debug(f"[Exchange] LDAP search base: {searchbase}")
        logging.debug(f"[Exchange] LDAP search filter: {search_filter}")
        
        return self.ldap_session.extend.standard.paged_search(
            searchbase, 
            search_filter, 
            attributes=properties, 
            paged_size=1000, 
            generator=True,
            controls=sd_flag,
            search_scope=search_scope,
            no_cache=no_cache,
            no_vuln_check=no_vuln_check,
            raw=raw
        )
    
    def get_exchange_groups(self, properties=None, searchbase=None, sd_flag=None,
                         search_scope=ldap3.SUBTREE, no_cache=False, no_vuln_check=False, raw=False):
        """
        Find Exchange security groups in the domain
        
        Args:
            properties: Properties to retrieve (optional)
            searchbase: Search base for the LDAP query (optional)
            search_scope: LDAP search scope (default: SUBTREE)
            no_cache: Whether to use cache (default: False)
            no_vuln_check: Whether to skip vulnerability checks (default: False)
            raw: Whether to return raw LDAP entries (default: False)
        Returns:
            List of Exchange security groups with their properties
        """
        if not properties:
            properties = [
                "name",
                "objectGUID", 
                "objectSid",
                "member",
                "groupType",
                "msExchGroupDepartRestriction",
                "msExchGroupJoinRestriction",
                "whenCreated",
                "whenChanged"
            ]
        
        if not searchbase:
            searchbase = self.configuration_dn
        
        if sd_flag:
            sd_flag = security_descriptor_control(sdflags=sd_flag)
        
        # This will find Exchange-related security groups
        search_filter = "(&(objectClass=group)(|(name=Exchange*)(name=*Mail*)(name=Organization Management)(name=*recipient*)))"
        
        logging.debug(f"[Exchange] LDAP search base: {searchbase}")
        logging.debug(f"[Exchange] LDAP search filter: {search_filter}")
        
        return self.ldap_session.extend.standard.paged_search(
            searchbase, 
            search_filter, 
            attributes=properties, 
            paged_size=1000, 
            generator=True,
            controls=sd_flag,
            search_scope=search_scope,
            no_cache=no_cache,
            no_vuln_check=no_vuln_check,
            raw=raw
        )
    
    def get_exchange_virtual_directories(self, properties=None, identity=None, searchbase=None, sd_flag=None,
                                      search_scope=ldap3.SUBTREE, no_cache=False, no_vuln_check=False, raw=False):
        """
        Find Exchange virtual directories in the domain
        
        Args:
            properties: Properties to retrieve (optional)
            identity: Directory name to target (optional)
            searchbase: Search base for the LDAP query (optional)
            search_scope: LDAP search scope (default: SUBTREE)
            no_cache: Whether to use cache (default: False)
            no_vuln_check: Whether to skip vulnerability checks (default: False)
            raw: Whether to return raw LDAP entries (default: False)
        Returns:
            List of Exchange virtual directories with their properties
        """
        if not properties:
            properties = [
                "name",
                "objectGUID",
                "msExchInternalHostname",
                "msExchExternalHostname",
                "msExchRequireSSL",
                "msExchUMVirtualDirectory",
                "whenCreated",
                "whenChanged"
            ]
        
        if not searchbase:
            searchbase = self.configuration_dn

        if sd_flag:
            sd_flag = security_descriptor_control(sdflags=sd_flag)
        
        # This will find various Exchange virtual directories (OWA, ECP, etc.)
        search_filter = "(objectClass=msExchVirtualDirectory)"
        
        if identity:
            search_filter = f"(&{search_filter}(name={identity}))"
        
        logging.debug(f"[Exchange] LDAP search base: {searchbase}")
        logging.debug(f"[Exchange] LDAP search filter: {search_filter}")
        
        return self.ldap_session.extend.standard.paged_search(
            searchbase, 
            search_filter, 
            attributes=properties, 
            paged_size=1000, 
            generator=True,
            controls=sd_flag,
            search_scope=search_scope,
            no_cache=no_cache,
            no_vuln_check=no_vuln_check,
            raw=raw
        )
    
    def get_exchange_ews_virtual_directories(self, properties=None, searchbase=None,
                                         search_scope=ldap3.SUBTREE, no_cache=False, no_vuln_check=False, raw=False):
        """
        Find Exchange Web Services (EWS) virtual directories
        
        Args:
            properties: Properties to retrieve (optional)
            searchbase: Search base for the LDAP query (optional)
            search_scope: LDAP search scope (default: SUBTREE)
            no_cache: Whether to use cache (default: False)
            no_vuln_check: Whether to skip vulnerability checks (default: False)
            raw: Whether to return raw LDAP entries (default: False)
        Returns:
            List of EWS virtual directories with their properties
        """
        if not properties:
            properties = [
                "name",
                "objectGUID",
                "msExchInternalHostname",
                "msExchExternalHostname",
                "msExchRequireSSL",
                "whenCreated",
                "whenChanged"
            ]
        
        if not searchbase:
            searchbase = self.configuration_dn
        
        search_filter = "(&(objectClass=msExchVirtualDirectory)(name=EWS*))"
        
        logging.debug(f"[Exchange] LDAP search base: {searchbase}")
        logging.debug(f"[Exchange] LDAP search filter: {search_filter}")
        
        return self.ldap_session.extend.standard.paged_search(
            searchbase, 
            search_filter, 
            attributes=properties, 
            paged_size=1000, 
            generator=True,
            controls=security_descriptor_control(sdflags=0x07),
            search_scope=search_scope,
            no_cache=no_cache,
            no_vuln_check=no_vuln_check,
            raw=raw
        )
    
    def analyze_exchange_permissions(self, objects, resolve_sids=True):
        """
        Analyze security permissions on Exchange objects to identify potential attack vectors
        
        Args:
            objects: List of Exchange LDAP objects to analyze
            resolve_sids: Whether to resolve SIDs to names (default: True)
            
        Returns:
            Dictionary of objects with their permissions and potential security issues
        """
        results = {}
        
        for obj in objects:
            if obj['type'] != 'searchResEntry':
                continue
                
            attrs = obj.get('attributes', {})
            obj_name = attrs.get('name', ['Unknown'])[0] if isinstance(attrs.get('name', []), list) else attrs.get('name', 'Unknown')
            
            # Determine object type
            obj_type = "Unknown"
            for cls, type_name in EXCHANGE_OBJECT_TYPES.items():
                if cls in str(attrs.get('objectClass', [])):
                    obj_type = type_name
                    break
            
            # Skip objects with no security descriptor
            if 'nTSecurityDescriptor' not in attrs:
                continue
            
            # Parse security descriptor
            security = ExchangeSecurity(attrs['nTSecurityDescriptor'])
            
            # Resolver function for SIDs
            def resolve_sid(sid):
                if not resolve_sids:
                    return sid
                
                # First check well-known SIDs
                if sid in WELL_KNOWN_SIDS:
                    return WELL_KNOWN_SIDS[sid]
                
                # Try to resolve via PowerView
                try:
                    return self.powerview.convertfrom_sid(sid)
                except:
                    return sid
            
            # Get all permission details
            permissions = security.get_ace_details(resolve_sid if resolve_sids else None)
            
            # Check for dangerous permissions
            vulnerabilities = security.check_dangerous_permissions()
            if resolve_sids:
                for vuln in vulnerabilities:
                    vuln["Principal"] = resolve_sid(vuln["SID"])
            
            # Store results
            results[obj_name] = {
                "ObjectType": obj_type,
                "Owner": security.get_owner_name(resolve_sid if resolve_sids else None),
                "Permissions": permissions,
                "Vulnerabilities": vulnerabilities,
                "ObjectGUID": attrs.get('objectGUID', 'Unknown')
            }
        
        return results
    
    def check_mailbox_permissions(self, mailboxes, resolve_sids=True):
        """
        Check mailbox permissions for specific scenarios like Send-As, Full Access, etc.
        
        Args:
            mailboxes: List of mailbox LDAP objects to analyze
            resolve_sids: Whether to resolve SIDs to names (default: True)
            
        Returns:
            Dictionary of mailboxes with their special permissions
        """
        results = {}
        
        for mailbox in mailboxes:
            if mailbox['type'] != 'searchResEntry':
                continue
                
            attrs = mailbox.get('attributes', {})
            mailbox_name = attrs.get('name', ['Unknown'])[0] if isinstance(attrs.get('name', []), list) else attrs.get('name', 'Unknown')
            mail = attrs.get('mail', ['Unknown'])[0] if isinstance(attrs.get('mail', []), list) else attrs.get('mail', 'Unknown')
            
            # Skip objects with no security descriptor
            if 'nTSecurityDescriptor' not in attrs:
                continue
            
            # Parse security descriptor
            security = ExchangeSecurity(attrs['nTSecurityDescriptor'])
            
            # Resolver function for SIDs
            def resolve_sid(sid):
                if not resolve_sids:
                    return sid
                
                # First check well-known SIDs
                if sid in WELL_KNOWN_SIDS:
                    return WELL_KNOWN_SIDS[sid]
                
                # Try to resolve via PowerView
                try:
                    return self.powerview.convertfrom_sid(sid)
                except:
                    return sid
            
            # Look for specific permissions
            send_as_rights = []
            full_access_rights = []
            
            for sid, ace_data in security.aces.items():
                principal_name = resolve_sid(sid) if resolve_sids else sid
                
                # Check for Send-As rights
                for right_id in ace_data["extended_rights"]:
                    if right_id == "1131f6aa-9c07-11d1-f79f-00c04fc2dcd2":  # Send-As right
                        send_as_rights.append(principal_name)
                
                # Check for Full Access rights (combination of rights)
                if (ace_data["rights"] & ACTIVE_DIRECTORY_RIGHTS.GENERIC_ALL) or \
                   (ace_data["rights"] & ACTIVE_DIRECTORY_RIGHTS.GENERIC_READ and \
                    ace_data["rights"] & ACTIVE_DIRECTORY_RIGHTS.GENERIC_WRITE):
                    full_access_rights.append(principal_name)
            
            # Store results
            results[mailbox_name] = {
                "Email": mail,
                "SendAsRights": list(set(send_as_rights)),  # Remove duplicates
                "FullAccessRights": list(set(full_access_rights)),  # Remove duplicates
                "ObjectGUID": attrs.get('objectGUID', 'Unknown')
            }
        
        return results
    
    def detect_exchange_version(self, exchange_objects):
        """
        Detect Exchange version from Exchange server objects
        
        Args:
            exchange_objects: List of Exchange server LDAP objects
            
        Returns:
            String describing detected Exchange version
        """
        versions = {}
        
        for obj in exchange_objects:
            if obj['type'] != 'searchResEntry':
                continue
                
            attrs = obj.get('attributes', {})
            
            # Check for serialNumber which often contains version info
            if 'serialNumber' in attrs:
                serial = attrs['serialNumber']
                if isinstance(serial, list):
                    serial = serial[0]
                
                if 'Version 15' in serial:
                    if 'CU' in serial:
                        match = re.search(r'Version 15\.(\d+)\.(\d+)\.(\d+)', serial)
                        if match:
                            versions["Exchange 2019/2016"] = serial
                    else:
                        versions["Exchange 2013"] = serial
                elif 'Version 14' in serial:
                    versions["Exchange 2010"] = serial
                elif 'Version 8' in serial:
                    versions["Exchange 2007"] = serial
            
            # Check for msExchProductID
            if 'msExchProductID' in attrs:
                prod_id = attrs['msExchProductID']
                if isinstance(prod_id, list):
                    prod_id = prod_id[0]
                
                # Map product IDs to versions if needed
                logging.debug(f"Exchange Product ID: {prod_id}")
        
        if not versions:
            return "Unknown Exchange version"
        
        # Return the newest version found
        if "Exchange 2019/2016" in versions:
            return versions["Exchange 2019/2016"]
        elif "Exchange 2013" in versions:
            return versions["Exchange 2013"]
        elif "Exchange 2010" in versions:
            return versions["Exchange 2010"]
        elif "Exchange 2007" in versions:
            return versions["Exchange 2007"]
        
        return "Unknown Exchange version" 


#!/usr/bin/env python3

from .helpers import in_exception, LdapObfuscate
from .validators import (
	is_sid_attribute, is_dn_attribute, is_bitwise_attribute,
	is_oid, randomly_prepend_zeros_oid, randomly_hex_encode_dn_string
) 


#!/usr/bin/env python3
import random
import re

from powerview.utils.helpers import is_valid_dn
from .constants import ATTRIBUTE_OID, WILDCARD
from .exceptions import LdapParserException
from ..utils.helpers import in_exception, LdapObfuscate
from ..utils.validators import (
	is_sid_attribute, is_dn_attribute, is_bitwise_attribute
)

class Operators:
	NOT = '!'
	AND = '&'
	OR = '|'

class LdapToken:
	def __init__(self, content, token_type):
		self.content = content
		self.type = token_type

	def __repr__(self):
		return f"{self.type}: {self.content}"

class AttributeParser:
	def __init__(self, attributes):
		if attributes is None:
			self.attributes = []
		elif isinstance(attributes, str):
			self.attributes = [attributes]
		elif isinstance(attributes, (list, tuple)):
			self.attributes = list(attributes)
		else:
			self.attributes = []

	def get_attributes(self):
		return self.attributes

	def random_oid(self):
		for i in range(len(self.attributes)):
			if in_exception(self.attributes[i]):
				continue
				
			if random.choice([True, False]):
				oid = ATTRIBUTE_OID.get(self.attributes[i])
				self.attributes[i] = oid if oid else self.attributes[i]

	def random_casing(self):
		for i in range(len(self.attributes)):
			if in_exception(self.attributes[i]):
				continue
				
			self.attributes[i] = LdapObfuscate.casing(self.attributes[i])

class DNParser:
	def __init__(self, dn):
		self.dn = str(dn) if dn is not None else ""
		self.enable_spacing = False
		self.parsed_structure = []

	def parse(self):
		if not self.dn:
			return []
		pattern = re.compile(r'([A-Za-z]+)=([^,]+)')
		matches = pattern.findall(self.dn)
		self.parsed_structure = [{"attribute": attr, "value": val} for attr, val in matches]
		return self.parsed_structure

	def dn_hex(self, parsed_structure=None):
		if parsed_structure is None:
			parsed_structure = self.parsed_structure

		for i in range(len(parsed_structure)):
			parsed_structure[i]["value"] = LdapObfuscate.randhex(parsed_structure[i]["value"])

	def dn_randomcase(self, parsed_structure=None):
		if parsed_structure is None:
			parsed_structure = self.parsed_structure

		for i in range(len(parsed_structure)):
			parsed_structure[i]["attribute"] = LdapObfuscate.casing(parsed_structure[i]["attribute"])
			parsed_structure[i]["value"] = LdapObfuscate.casing(parsed_structure[i]["value"])

	def convert_to_dn(self, parsed_structure=None):
		if parsed_structure is None:
			parsed_structure = self.parsed_structure

		if not parsed_structure:
			return ""
		return ','.join([f"{item['attribute']}={item['value']}" for item in parsed_structure])

	def random_spacing(self):
		self.enable_spacing = True

class FilterParser:
	"""
	LDAP Filter Parser class to parse LDAP filters and obfuscate them
	
	Reference: https://i.blackhat.com/BH-US-24/Presentations/US24-Bohannon-MaLDAPtive-Diving-Deep-Into-LDAP-Wednesday.pdf
	"""
	TOKEN_PATTERNS = {
		'group_start': re.compile(r'\('),
		'group_end': re.compile(r'\)'),
		'boolean_operator': re.compile(r'[&|!]'),
		'comparison_operator': re.compile(r'([<>]?=|:=)'),
		'attribute': re.compile(r'([a-zA-Z0-9-]+)'),
		'extensible_match': re.compile(r':([0-9.]+):'),
		'value': re.compile(r'([^)]*?)(?=\)|$)')
	}

	def __init__(self, ldap_filter):
		if not ldap_filter or not isinstance(ldap_filter, str):
			raise LdapParserException("Invalid LDAP filter: must be a non-empty string")
		self.ldap_filter = ldap_filter.strip()
		self.tokens = []
		self.parsed_structure = []
		self.enable_spacing = False

	def get_parsed_structure(self):
		return self.parsed_structure

	def _is_sid_attribute(self, attr):
		"""Check if attribute typically contains SID values"""
		return is_sid_attribute(attr)

	def _is_dn_attribute(self, attr):
		"""Check if attribute typically contains DN values"""
		return is_dn_attribute(attr)

	def _is_bitwise_attribute(self, attr):
		"""Check if attribute supports bitwise operations"""
		return is_bitwise_attribute(attr)

	def _apply_to_filter_leaves(self, parsed_structure, transform_func):
		"""
		Apply transformation function to leaf filter elements (attribute-operator-value triplets).
		Following Go LeafApplierFilterMiddleware pattern.
		"""
		if not parsed_structure:
			return

		for i in range(len(parsed_structure)):
			if isinstance(parsed_structure[i], list):
				self._apply_to_filter_leaves(parsed_structure[i], transform_func)
			elif (isinstance(parsed_structure[i], dict) and 
				  parsed_structure[i].get("type") == "Attribute" and
				  i + 2 < len(parsed_structure) and
				  isinstance(parsed_structure[i+1], dict) and
				  parsed_structure[i+1].get("type") == "ComparisonOperator" and
				  isinstance(parsed_structure[i+2], dict) and
				  parsed_structure[i+2].get("type") == "Value"):
				
				attr = parsed_structure[i].get("content", "")
				operator = parsed_structure[i+1].get("content", "")
				value = parsed_structure[i+2].get("content", "")
				
				new_attr, new_value = transform_func(attr, operator, value)
				
				if new_attr != attr:
					parsed_structure[i]["content"] = new_attr
				if new_value != value:
					parsed_structure[i+2]["content"] = new_value

	def parse(self):
		try:
			self.tokenize()
			self.parsed_structure = self.build_filter_structure()
			return self.parsed_structure
		except Exception as e:
			raise LdapParserException(f"Failed to parse LDAP filter: {str(e)}")

	def handle_escaped_chars(self, value):
		"""Handle LDAP escaped characters properly"""
		if not value:
			return value
		value = value.replace('\\)', ')')
		value = value.replace('\\(', '(')
		value = value.replace('\\*', '*')
		value = value.replace('\\\\', '\\')
		return value

	def tokenize(self):
		cursor = 0
		max_iterations = len(self.ldap_filter) * 2
		iterations = 0
		
		while cursor < len(self.ldap_filter) and iterations < max_iterations:
			iterations += 1
			
			if cursor >= len(self.ldap_filter):
				break
				
			char = self.ldap_filter[cursor]

			if self.TOKEN_PATTERNS['group_start'].match(char):
				self.tokens.append(LdapToken(char, 'GroupStart'))
				cursor += 1
			elif self.TOKEN_PATTERNS['group_end'].match(char):
				self.tokens.append(LdapToken(char, 'GroupEnd'))
				cursor += 1
			elif self.TOKEN_PATTERNS['boolean_operator'].match(char):
				self.tokens.append(LdapToken(char, 'BooleanOperator'))
				cursor += 1
			else:
				attribute_match = self.TOKEN_PATTERNS['attribute'].match(self.ldap_filter[cursor:])
				if attribute_match:
					attribute = attribute_match.group(1)
					cursor += len(attribute)

					extensible_match = self.TOKEN_PATTERNS['extensible_match'].match(self.ldap_filter[cursor:])
					if extensible_match:
						oid = extensible_match.group(1)
						self.tokens.append(LdapToken(attribute, 'Attribute'))
						self.tokens.append(LdapToken(oid, 'ExtensibleMatchFilter'))
						cursor += len(extensible_match.group(0))

						comparison_match = self.TOKEN_PATTERNS['comparison_operator'].match(self.ldap_filter[cursor:])
						if comparison_match:
							comparison = comparison_match.group(1)
							self.tokens.append(LdapToken(comparison, 'ComparisonOperator'))
							cursor += len(comparison)

							value_match = self.TOKEN_PATTERNS['value'].match(self.ldap_filter[cursor:])
							if value_match:
								value = value_match.group(1)
								value = self.handle_escaped_chars(value)
								self.tokens.append(LdapToken(value, 'Value'))
								cursor += len(value_match.group(1))
							else:
								raise LdapParserException(f"Malformed LDAP filter: value missing after {attribute}")
					else:
						self.tokens.append(LdapToken(attribute, 'Attribute'))

						comparison_match = self.TOKEN_PATTERNS['comparison_operator'].match(self.ldap_filter[cursor:])
						if comparison_match:
							comparison = comparison_match.group(1)
							self.tokens.append(LdapToken(comparison, 'ComparisonOperator'))
							cursor += len(comparison)

							value_match = self.TOKEN_PATTERNS['value'].match(self.ldap_filter[cursor:])
							if value_match:
								value = value_match.group(1)
								value = self.handle_escaped_chars(value)
								self.tokens.append(LdapToken(value, 'Value'))
								cursor += len(value_match.group(1))
							else:
								raise LdapParserException(f"Malformed LDAP filter: value missing after {attribute}")
					continue
				else:
					if char.isspace():
						cursor += 1
						continue
					else:
						raise LdapParserException(f"Unexpected character '{char}' at position {cursor}")

	def build_filter_structure(self):
		if not self.tokens:
			return []
			
		stack = []
		current_filter = []

		for token in self.tokens:
			if token.type == 'GroupStart':
				stack.append(current_filter)
				current_filter = []
			elif token.type == 'GroupEnd':
				if not stack:
					raise LdapParserException("Malformed LDAP filter: unmatched closing parenthesis")
				last_filter = current_filter
				current_filter = stack.pop()
				current_filter.append(last_filter)
			else:
				current_filter.append({
					"type": token.type,
					"content": token.content
				})

		if stack:
			raise LdapParserException("Malformed LDAP filter: unmatched opening parenthesis")

		return current_filter

	def random_spacing(self):
		self.enable_spacing = True

	def convert_to_ldap(self, parsed_structure=None):
		if parsed_structure is None:
			parsed_structure = self.parsed_structure

		if not parsed_structure:
			return ""

		ldap_string = []
		skip_random_spacing = False

		def get_last_char(string_list):
			for s in reversed(string_list):
				if s and isinstance(s, str) and s:
					return s[-1]
			return None

		for i in range(len(parsed_structure)):
			token = parsed_structure[i]
			if isinstance(token, list):
				prev_char = get_last_char(ldap_string)
				spacing_before = "" if not ldap_string else LdapObfuscate.get_context_aware_spacing(prev_char, '(')
				
				nested_result = self.convert_to_ldap(token)
				if nested_result:
					ldap_string.append(f"{spacing_before}({nested_result})")
					
					if i < len(parsed_structure) - 1:
						next_token = parsed_structure[i+1]
						next_char = '(' if isinstance(next_token, list) else (
							next_token.get('content', '')[0] if next_token.get('content') else None
						)
						spacing_after = LdapObfuscate.get_context_aware_spacing(')', next_char)
						ldap_string.append(spacing_after)
			else:
				if token["type"] == "Attribute":
					if in_exception(token["content"]) or (i+1 < len(parsed_structure) and 
					   parsed_structure[i+1].get("type") == "ExtensibleMatchFilter"):
						skip_random_spacing = True
					else:
						skip_random_spacing = False

					prev_char = get_last_char(ldap_string)
					next_char = (parsed_structure[i+1].get("content", "")[0] 
							   if i+1 < len(parsed_structure) and parsed_structure[i+1].get("content") 
							   else None)
					spacing = "" if not ldap_string else LdapObfuscate.get_context_aware_spacing(prev_char, next_char)
					
					if token.get('content'):
						ldap_string.append(f"{spacing}{token['content']}")

				elif token["type"] == "ComparisonOperator":
					spacing = "" if skip_random_spacing else LdapObfuscate.random_spaces(1, 2)
					if token.get('content'):
						ldap_string.append(f"{spacing}{token['content']}")

				elif token["type"] == "Value":
					content = token.get('content', '')
					spacing = "" if skip_random_spacing else LdapObfuscate.get_context_aware_spacing(
						'=', 
						content[0] if content else None
					)
					ldap_string.append(f"{spacing}{content}")

				elif token["type"] == "ExtensibleMatchFilter":
					if token.get('content'):
						ldap_string.append(f":{token['content']}:")

				else:
					spacing = "" if skip_random_spacing or not ldap_string else LdapObfuscate.random_spaces(0, 2)
					if token.get('content'):
						ldap_string.append(f"{spacing}{token['content']}")

		result = ''.join(s for s in ldap_string if s is not None)
		return result if result else "" 


"""
MCP Server implementation for PowerView.
"""

import asyncio
import logging
import json
import threading
import sys
from typing import Dict, List, Optional, Any, Tuple, Union, Callable

from .src import tools, prompts

try:
    from mcp.server.fastmcp import FastMCP, Context
    import mcp.types as types
    MCP_AVAILABLE = True
except ImportError:
    MCP_AVAILABLE = False

class MCPServer:
    """
    Model Context Protocol server for PowerView.
    
    This class implements an MCP server that exposes PowerView's functionality
    to AI assistants using the Model Context Protocol.
    """
    
    def __init__(self, powerview, name="PowerView MCP", host="127.0.0.1", port=8080):
        """
        Initialize the MCP server.
        
        Args:
            powerview: PowerView instance to expose via MCP
            name: Name of the MCP server
            host: Host to bind the server to
            port: Port to bind the server to
        """
        if not MCP_AVAILABLE:
            raise ImportError("MCP dependencies not installed. Install with: pip install .[mcp]")
            
        self.powerview = powerview
        self.name = name
        self.host = host
        self.port = port
        self.mcp = FastMCP(self.name)
        self.status = False
        self.server_thread = None
        # self._setup_resources() # Remove or comment out this line
        tools.setup_tools(self.mcp, self.powerview)
        prompts.setup_prompts(self.mcp)

    def set_status(self, status):
        self.status = status

    def get_status(self):
        return self.status

    async def _server_started(self):
        """Callback that runs when the server is actually ready to accept connections"""
        self.set_status(True)
        logging.info("MCP server is ready to accept connections")

    def start(self):
        """Start the MCP server."""
        if self.server_thread and self.server_thread.is_alive():
            logging.warning("MCP server is already running")
            return
        
        def run_server():
            import uvicorn
            
            logging.info(f"Starting MCP server on {self.host}:{self.port}")
            try:
                # Create an ASGI application from the MCP server
                app = self.mcp.sse_app()
                
                # Set status before starting server
                self.set_status(True)
                
                # Start the server with uvicorn
                uvicorn.run(
                    app=app,
                    host=self.host,
                    port=self.port,
                    log_level="error",
                    access_log=False
                )
            except Exception as e:
                self.set_status(False)
                logging.error(f"Error starting MCP server: {str(e)}")
                sys.exit(1)

        # Create and start the server thread
        self.server_thread = threading.Thread(target=run_server, daemon=True)
        self.server_thread.start()
        
        import time
        time.sleep(0.2)
        
        logging.debug(f"MCP server thread started, status: {self.get_status()}")

    def stop(self):
        """Stop the MCP server."""
        self.set_status(False)
        logging.info("Stopping MCP server...")
        
        # The MCP server will stop when the main thread exits since we use a daemon thread 


#!/usr/bin/env python3
# original script: https://gist.github.com/GeisericII/6849bc86620c7a764d88502df5187bd0

import logging
import re
import time
from impacket.dcerpc.v5 import transport, rrp
from impacket.smbconnection import SessionError

class RemoteOperations:
	KNOWN_PROTOCOLS = {
        139: {'bindstr': r'ncacn_np:%s[\pipe\winreg]', 'set_host': True},
        445: {'bindstr': r'ncacn_np:%s[\pipe\winreg]', 'set_host': True},
    }

	def __init__(self, connection, port=445):
		self.connection = connection
		self.port = port
		self.pipetriggered = False

	def connect(self, target):
		stringBinding = self.KNOWN_PROTOCOLS[self.port]['bindstr'] % target
		try:
			dce = self.connection.connectRPCTransport(host=target, stringBindings=stringBinding, interface_uuid=rrp.MSRPC_UUID_RRP, raise_exceptions=True)
			return dce
		except SessionError as e:
			if str(e).find('STATUS_PIPE_NOT_AVAILABLE') >= 0:
				logging.warning("Trying to start the Remote Registry...")
				time.sleep(1)
				if not self.pipetriggered:
					return self.connect(target)
				else:
					logging.error("Failed to bind")
			else:
				logging.error(str(e))
				return

	def query_logged_on(self, dce):
		sidRegex = "^S-1-5-21-[0-9]+-[0-9]+-[0-9]+-[0-9]+$"
		resp = rrp.hOpenUsers(dce)
		hKey = resp['phKey']
		users = list()
		index = 1
		while True:
			try:
				resp = rrp.hBaseRegEnumKey(dce, hKey, index)
				userSid = resp['lpNameOut'][:-1]
				res = re.match(sidRegex, userSid)
				if res:
					users.append(userSid)
				index += 1
			except Exception as e:
				break

		rrp.hBaseRegCloseKey(dce, hKey)
		dce.disconnect()

		return users


#!/usr/bin/env python3

from .attribute_obfuscation import (
	oid_attribute_obfuscation, anr_attribute_obfuscation, random_casing_obfuscation
)
from .value_obfuscation import (
	prepend_zeros_obfuscation, hex_value_obfuscation, spacing_obfuscation
)
from .operator_obfuscation import equality_to_approximation_obfuscation
from .wildcard_obfuscation import wildcard_expansion_obfuscation 


#!/usr/bin/env python3
import sys
try:
    sys.modules.pop('readline', None)
    import gnureadline as readline
    sys.modules['readline'] = readline
except ImportError:
    import readline
    sys.modules['readline'] = readline
from powerview.powerview import PowerView
from powerview.utils.helpers import *
from powerview.utils.native import *
from powerview.utils.formatter import FORMATTER
from powerview.utils.completer import Completer
from powerview.utils.connections import CONNECTION
from powerview.utils.logging import LOG
from powerview.utils.parsers import powerview_arg_parse, arg_parse
from powerview.utils.shell import get_prompt
from powerview.utils.colors import bcolors, Gradient
from powerview.utils.history import get_shell_history

import ldap3
import random
import string
import shlex

def main():
    """
    Main entry point for PowerView tool.
    
    Handles command-line argument parsing, LDAP connection setup,
    and interactive command processing.
    """
    args = arg_parse()

    flat_domain = args.domain.split('.')[0] if '.' in args.domain else args.domain
    flat_domain = sanitize_component(flat_domain.lower())
    username = sanitize_component(args.username.lower())
    ldap_address = sanitize_component(args.ldap_address.lower())

    components = [flat_domain, username, ldap_address]
    folder_name = '-'.join(filter(None, components)) or "default-log"

    log_handler = LOG(folder_name)

    if args.debug:
        logging = log_handler.setup_logger("DEBUG")
    else:
        logging = log_handler.setup_logger()
    
    try:
        conn = CONNECTION(args)
        init_ldap_address = args.ldap_address
        powerview = PowerView(conn, args)
        if powerview.ldap_session and powerview.ldap_session.bound:
            powerview.add_domain_connection(powerview.conn.domain)

        comp = Completer()
        comp.setup_completer()

        current_target_domain = None

        using_cache = False

        while True:
            try:
                temp_powerview = None
                if args.query:
                    cmd = args.query
                else:
                    cmd = input(get_prompt(powerview, current_target_domain, using_cache, args))

                if cmd:
                    try:
                        cmd = shlex.split(cmd)
                    except ValueError as e:
                        if args.stack_trace:
                            raise e
                        else:
                            logging.error(str(e))
                            continue

                    pv_args = powerview_arg_parse(cmd)

                    if pv_args:
                        if pv_args.server and pv_args.server.lower() != powerview.domain.lower():
                            try:
                                temp_powerview = powerview.get_domain_powerview(pv_args.server)
                                current_target_domain = pv_args.server
                            except ldap3.core.exceptions.LDAPSocketOpenError as e:
                                logging.error(f'Connection to domain {pv_args.server} failed: {str(e)}')
                                current_target_domain = None
                                continue
                            except ldap3.core.exceptions.LDAPBindError as e:
                                logging.error(f'Authentication to domain {pv_args.server} failed: {str(e)}')
                                current_target_domain = None
                                continue
                            except Exception as e:
                                logging.error(f'Domain {pv_args.server} operation failed: {str(e)}')
                                current_target_domain = None
                                if args.stack_trace:
                                    import traceback
                                    logging.debug(traceback.format_exc())
                                continue
                        else:
                            # No server specified or same as current domain
                            current_target_domain = None
                            temp_powerview = None
                            
                        try:
                            entries = None
                            if pv_args.module.casefold() == 'get-domain':
                                if temp_powerview:
                                    entries = temp_powerview.get_domain(args=pv_args)
                                else:
                                    entries = powerview.get_domain(args=pv_args)
                            elif pv_args.module.casefold() == 'get-domainobject' or pv_args.module.casefold() == 'get-adobject':
                                if temp_powerview:
                                    entries = temp_powerview.get_domainobject(args=pv_args)
                                else:
                                    entries = powerview.get_domainobject(args=pv_args)
                            elif pv_args.module.casefold() == 'get-domainobjectowner' or pv_args.module.casefold() == 'get-objectowner':
                                if temp_powerview:
                                    entries = temp_powerview.get_domainobjectowner(args=pv_args)
                                else:
                                    entries = powerview.get_domainobjectowner(args=pv_args)
                            elif pv_args.module.casefold() == 'get-domainobjectacl' or pv_args.module.casefold() == 'get-objectacl':
                                if temp_powerview:
                                    entries = temp_powerview.get_domainobjectacl(
                                        args=pv_args
                                    )
                                else:
                                    entries = powerview.get_domainobjectacl(
                                        args=pv_args
                                    )
                            elif pv_args.module.casefold() == 'get-domainuser':
                                if temp_powerview:
                                    entries = temp_powerview.get_domainuser(args=pv_args)
                                else:
                                    entries = powerview.get_domainuser(args=pv_args)
                            elif pv_args.module.casefold() == 'get-localuser':
                                properties = pv_args.properties if pv_args.properties else None
                                computername = pv_args.computer if pv_args.computer else pv_args.computername
                                if temp_powerview:
                                    entries = temp_powerview.get_localuser(computer_name=computername, identity=pv_args.identity, properties=properties, args=pv_args)
                                else:
                                    entries = powerview.get_localuser(computer_name=computername, identity=pv_args.identity, properties=properties, args=pv_args)
                            elif pv_args.module.casefold() == 'get-domaincomputer':
                                if temp_powerview:
                                    entries = temp_powerview.get_domaincomputer(args=pv_args)
                                else:
                                    entries = powerview.get_domaincomputer(args=pv_args)
                            elif pv_args.module.casefold() == 'get-domaingroup':
                                properties = pv_args.properties if pv_args.properties else None
                                identity = pv_args.identity.strip() if pv_args.identity else None
                                if temp_powerview:
                                    entries = temp_powerview.get_domaingroup(pv_args, properties, identity, no_cache=pv_args.no_cache)
                                else:
                                    entries = powerview.get_domaingroup(pv_args, properties, identity, no_cache=pv_args.no_cache)
                            elif pv_args.module.casefold() == 'get-domaingroupmember':
                                identity = pv_args.identity.strip()
                                if temp_powerview:
                                    entries = temp_powerview.get_domaingroupmember(identity=identity, args=pv_args)
                                else:
                                    entries = powerview.get_domaingroupmember(identity=identity, args=pv_args)
                            elif pv_args.module.casefold() == 'get-domainforeigngroupmember' or pv_args.module.casefold() == 'find-foreigngroup':
                                if temp_powerview:
                                    entries = temp_powerview.get_domainforeigngroupmember(pv_args)
                                else:
                                    entries = powerview.get_domainforeigngroupmember(pv_args)
                            elif pv_args.module.casefold() == 'get-domainforeignuser' or pv_args.module.casefold() == 'find-foreignuser':
                                if temp_powerview:
                                    entries = temp_powerview.get_domainforeignuser(pv_args)
                                else:
                                    entries = powerview.get_domainforeignuser(pv_args)
                            elif pv_args.module.casefold() == 'get-domaincontroller':
                                if temp_powerview:
                                    entries = temp_powerview.get_domaincontroller(args=pv_args)
                                else:
                                    entries = powerview.get_domaincontroller(args=pv_args)
                            elif pv_args.module.casefold() == 'get-domaingpo':
                                properties = pv_args.properties if pv_args.properties else None
                                identity = pv_args.identity.strip() if pv_args.identity else None
                                if temp_powerview:
                                    entries = temp_powerview.get_domaingpo(args=pv_args)
                                else:
                                    entries = powerview.get_domaingpo(args=pv_args)
                            elif pv_args.module.casefold() == 'get-domaingpolocalgroup' or pv_args.module.casefold() == 'get-gpolocalgroup':
                                if temp_powerview:
                                    entries = temp_powerview.get_domaingpolocalgroup(args=pv_args)
                                else:
                                    entries = powerview.get_domaingpolocalgroup(args=pv_args)
                            elif pv_args.module.casefold() == 'get-domaingposettings' or pv_args.module.casefold() == 'get-gposettings':
                                if temp_powerview:
                                    entries = temp_powerview.get_domaingposettings(args=pv_args)
                                else:
                                    entries = powerview.get_domaingposettings(args=pv_args)
                            elif pv_args.module.casefold() == 'get-domainou':
                                if temp_powerview:
                                    entries = temp_powerview.get_domainou(args=pv_args)
                                else:
                                    entries = powerview.get_domainou(args=pv_args)
                            elif pv_args.module.casefold() == 'get-domaindnszone':
                                identity = pv_args.identity.strip() if pv_args.identity else None
                                properties = pv_args.properties if pv_args.properties else None
                                if temp_powerview:
                                    entries = temp_powerview.get_domaindnszone(identity, properties, args=pv_args)
                                else:
                                    entries = powerview.get_domaindnszone(identity, properties, args=pv_args)
                            elif pv_args.module.casefold() == 'get-domaindnsrecord':
                                zonename = pv_args.zonename.strip() if pv_args.zonename else None
                                identity = pv_args.identity.strip() if pv_args.identity else None
                                properties = pv_args.properties if pv_args.properties else None
                                if temp_powerview:
                                    entries = temp_powerview.get_domaindnsrecord(identity, zonename, properties, args=pv_args)
                                else:
                                    entries = powerview.get_domaindnsrecord(identity, zonename, properties, args=pv_args)
                            elif pv_args.module.casefold() == 'get-domainsccm' or pv_args.module.casefold() == 'get-sccm':
                                properties = pv_args.properties if pv_args.properties else None
                                identity = pv_args.identity.strip() if pv_args.identity else None
                                if temp_powerview:
                                    entries = temp_powerview.get_domainsccm(pv_args, properties, identity)
                                else:
                                    entries = powerview.get_domainsccm(pv_args, properties, identity)
                            elif pv_args.module.casefold() == 'get-domaingmsa' or pv_args.module.casefold() == 'get-gmsa':
                                if temp_powerview:
                                    entries = temp_powerview.get_domaingmsa(args=pv_args)
                                else:
                                    entries = powerview.get_domaingmsa(args=pv_args)
                            elif pv_args.module.casefold() == 'get-domaindmsa' or pv_args.module.casefold() == 'get-dmsa':
                                if temp_powerview:
                                    entries = temp_powerview.get_domaindmsa(args=pv_args)
                                else:
                                    entries = powerview.get_domaindmsa(args=pv_args)
                            elif pv_args.module.casefold() == 'get-domainrbcd' or pv_args.module.casefold() == 'get-rbcd':
                                if temp_powerview:
                                    entries = temp_powerview.get_domainrbcd(args=pv_args)
                                else:
                                    entries = powerview.get_domainrbcd(args=pv_args)
                            elif pv_args.module.casefold() == 'get-domainwds' or pv_args.module.casefold() == 'get-wds':
                                if temp_powerview:
                                    entries = temp_powerview.get_domainwds(args=pv_args)
                                else:
                                    entries = powerview.get_domainwds(args=pv_args)
                            elif pv_args.module.casefold() == 'get-domainca' or pv_args.module.casefold() == 'get-ca':
                                if temp_powerview:
                                    entries = temp_powerview.get_domainca(args=pv_args)
                                else:
                                    entries = powerview.get_domainca(args=pv_args)
                            elif pv_args.module.casefold() == 'get-domaincatemplate' or pv_args.module.casefold() == 'get-catemplate':
                                properties = pv_args.properties if pv_args.properties else None
                                identity = pv_args.identity.strip() if pv_args.identity else None
                                if temp_powerview:
                                    entries = temp_powerview.get_domaincatemplate(pv_args, properties, identity)
                                else:
                                    entries = powerview.get_domaincatemplate(pv_args, properties, identity)
                            elif pv_args.module.casefold() == 'remove-domaincatemplate' or pv_args.module.casefold() == 'remove-catemplate':
                                if not pv_args.identity:
                                    logging.error("-Identity flag is required")
                                    continue

                                if temp_powerview:
                                    temp_powerview.remove_domaincatemplate(identity=pv_args.identity, args=pv_args)
                                else:
                                    powerview.remove_domaincatemplate(identity=pv_args.identity, args=pv_args)
                            elif pv_args.module.casefold() == 'add-domaincatemplate' or pv_args.module.casefold() == 'add-catemplate':
                                if pv_args.displayname is None:
                                    logging.error("-DisplayName flag is required")
                                    continue

                                displayname = pv_args.displayname
                                name = pv_args.name
                                if temp_powerview:
                                    temp_powerview.add_domaincatemplate(displayname, name, args=pv_args)
                                else:
                                    powerview.add_domaincatemplate(displayname, name, args=pv_args)
                            elif pv_args.module.casefold() == 'add-domaincatemplateacl' or pv_args.module.casefold() == 'add-catemplateacl':
                                if pv_args.template is not None and pv_args.principalidentity is not None and pv_args.rights is not None:
                                    if temp_powerview:
                                        temp_powerview.add_domaincatemplateacl(pv_args.template, pv_args.principalidentity, args=pv_args)
                                    else:
                                        powerview.add_domaincatemplateacl(pv_args.template, pv_args.principalidentity, args=pv_args)
                                else:
                                    logging.error('-TargetIdentity , -PrincipalIdentity and -Rights flags are required')
                            elif pv_args.module.casefold() == 'get-domaintrust':
                                properties = pv_args.properties if pv_args.properties else None
                                identity = pv_args.identity.strip() if pv_args.identity else None
                                if temp_powerview:
                                    entries = temp_powerview.get_domaintrust(args=pv_args)
                                else:
                                    entries = powerview.get_domaintrust(args=pv_args)
                            elif pv_args.module.casefold() == 'get-domaintrustkey' or pv_args.module.casefold() == 'get-trustkey':
                                if temp_powerview:
                                    entries = temp_powerview.get_domaintrustkey(args=pv_args)
                                else:
                                    entries = powerview.get_domaintrustkey(args=pv_args)
                            elif pv_args.module.casefold() == 'convertfrom-uacvalue':
                                if pv_args.value:
                                    value = pv_args.value.strip()
                                    if temp_powerview:
                                        entries = temp_powerview.convertfrom_uacvalue(value=value, output=True)
                                    else:
                                        entries = powerview.convertfrom_uacvalue(value=value, output=True)
                                else:
                                    logging.error("-Value flag is required")
                            elif pv_args.module.casefold() == 'convertfrom-sid':
                                if pv_args.objectsid:
                                    objectsid = pv_args.objectsid.strip()
                                    if temp_powerview:
                                        temp_powerview.convertfrom_sid(objectsid=objectsid, output=True, no_cache=pv_args.no_cache)
                                    else:
                                        powerview.convertfrom_sid(objectsid=objectsid, output=True, no_cache=pv_args.no_cache)
                                else:
                                    logging.error("-ObjectSID flag is required")
                            elif pv_args.module.casefold() == 'clear-cache':
                                if temp_powerview:
                                    temp_powerview.clear_cache()
                                else:
                                    powerview.clear_cache()
                                using_cache = False
                            elif pv_args.module.casefold() == 'login-as':
                                if temp_powerview:
                                    powerview.login_as(args=pv_args)
                                else:
                                    powerview.login_as(args=pv_args)
                            elif pv_args.module.casefold() == 'get-namedpipes':
                                if pv_args.computer is not None or pv_args.computername is not None:
                                    if temp_powerview:
                                        entries = temp_powerview.get_namedpipes(
                                            pv_args, 
                                            timeout=pv_args.timeout, 
                                            max_threads=pv_args.max_threads
                                        )
                                    else:
                                        entries = powerview.get_namedpipes(
                                            pv_args, 
                                            timeout=pv_args.timeout, 
                                            max_threads=pv_args.max_threads
                                        )
                                else:
                                    logging.error('-Computer or -ComputerName is required')
                            elif pv_args.module.casefold() == 'get-netshare':
                                if pv_args.computer is not None or pv_args.computername is not None:
                                    if temp_powerview:
                                        entries =  temp_powerview.get_netshare(pv_args)
                                    else:
                                        entries = powerview.get_netshare(pv_args)
                                else:
                                    logging.error('-Computer or -ComputerName is required')
                            elif pv_args.module.casefold() == 'get-regloggedon':
                                if pv_args.computer is not None or pv_args.computername is not None:
                                    computername = pv_args.computer if pv_args.computer else pv_args.computername
                                    if temp_powerview:
                                        entries = temp_powerview.get_regloggedon(computer_name=computername, args=pv_args)
                                    else:
                                        entries = powerview.get_regloggedon(computer_name=computername, args=pv_args)
                                else:
                                    logging.error('-Computer or -ComputerName is required')
                            elif pv_args.module.casefold() == 'get-netcomputerinfo':
                                if pv_args.computer is not None or pv_args.computername is not None:
                                    computername = pv_args.computer if pv_args.computer else pv_args.computername
                                    if temp_powerview:
                                        entries = temp_powerview.get_netcomputerinfo(computer_name=computername, args=pv_args)
                                    else:
                                        entries = powerview.get_netcomputerinfo(computer_name=computername, args=pv_args)
                                else:
                                    logging.error('-Computer or -ComputerName is required')
                            elif pv_args.module.casefold() == 'get-netloggedon':
                                if pv_args.computer is not None or pv_args.computername is not None:
                                    computername = pv_args.computer if pv_args.computer else pv_args.computername
                                    if temp_powerview:
                                        entries = temp_powerview.get_netloggedon(computer_name=computername, args=pv_args)
                                    else:
                                        entries = powerview.get_netloggedon(computer_name=computername, args=pv_args)
                                else:
                                    logging.error('-Computer or -ComputerName is required')
                            elif pv_args.module.casefold() == 'get-netservice':
                                if pv_args.computer is not None or pv_args.computername is not None:
                                    computername = pv_args.computer if pv_args.computer else pv_args.computername
                                    if temp_powerview:
                                        entries = temp_powerview.get_netservice(
                                            computer_name=computername,
                                            name=pv_args.name,
                                            is_running=pv_args.isrunning,
                                            is_stopped=pv_args.isstopped
                                        )
                                    else:
                                        entries = powerview.get_netservice(
                                            computer_name=computername,
                                            name=pv_args.name,
                                            is_running=pv_args.isrunning,
                                            is_stopped=pv_args.isstopped
                                        )
                                else:
                                    logging.error('-Computer or -ComputerName is required')
                            elif pv_args.module.casefold() == 'set-netservice':
                                if pv_args.computer is not None:
                                    if temp_powerview:
                                        succeed = temp_powerview.set_netservice(
                                            computer_name=pv_args.computer,
                                            service_name=pv_args.service_name,
                                            display_name=pv_args.display_name,
                                            binary_path=pv_args.binary_path,
                                            service_type=pv_args.service_type,
                                            start_type=pv_args.start_type,
                                            error_control=pv_args.error_control,
                                            service_start_name=pv_args.service_start_name,
                                            password=pv_args.password
                                        )
                                    else:
                                        succeed = powerview.set_netservice(
                                            computer_name=pv_args.computer,
                                            service_name=pv_args.service_name,
                                            display_name=pv_args.display_name,
                                            binary_path=pv_args.binary_path,
                                            service_type=pv_args.service_type,
                                            start_type=pv_args.start_type,
                                            error_control=pv_args.error_control,
                                            service_start_name=pv_args.service_start_name,
                                            password=pv_args.password
                                        )
                                else:
                                    logging.error('-Computer is required')
                            elif pv_args.module.casefold() == 'start-netservice':
                                if temp_powerview:
                                    succeed = temp_powerview.start_netservice(computer_name=pv_args.computer, service_name=pv_args.service_name)
                                else:
                                    succeed = powerview.start_netservice(computer_name=pv_args.computer, service_name=pv_args.service_name)
                            elif pv_args.module.casefold() == 'stop-netservice':
                                if temp_powerview:
                                    succeed = temp_powerview.stop_netservice(computer_name=pv_args.computer, service_name=pv_args.service_name)
                                else:
                                    succeed = powerview.stop_netservice(computer_name=pv_args.computer, service_name=pv_args.service_name)
                            elif pv_args.module.casefold() == 'get-netterminalsession' or pv_args.module.casefold() == 'qwinsta':
                                if pv_args.computer is not None or pv_args.computername is not None:
                                    computername = pv_args.computer if pv_args.computer else pv_args.computername
                                    if temp_powerview:
                                        entries = temp_powerview.get_netterminalsession(identity=computername, port=445, args=pv_args)
                                    else:
                                        entries = powerview.get_netterminalsession(identity=computername, port=445, args=pv_args)
                                else:
                                    logging.error('-Computer or -ComputerName is required')
                            elif pv_args.module.casefold() == 'remove-netterminalsession':
                                if pv_args.computer is not None or pv_args.computername is not None:
                                    computername = pv_args.computer if pv_args.computer else pv_args.computername
                                    if temp_powerview:
                                        succeed = temp_powerview.remove_netterminalsession(identity=computername, port=445, args=pv_args)
                                    else:
                                        succeed = powerview.remove_netterminalsession(identity=computername, port=445, args=pv_args)
                                else:
                                    logging.error('-Computer or -ComputerName is required')
                            elif pv_args.module.casefold() == 'stop-computer' or pv_args.module.casefold() == 'shutdown-computer':
                                if pv_args.computer is not None or pv_args.computername is not None:
                                    computername = pv_args.computer if pv_args.computer else pv_args.computername
                                    if temp_powerview:
                                        succeed = temp_powerview.stop_computer(identity=computername, port=445, args=pv_args)
                                    else:
                                        succeed = powerview.stop_computer(identity=computername, port=445, args=pv_args)
                                else:
                                    logging.error('-Computer or -ComputerName is required')
                            elif pv_args.module.casefold() == 'restart-computer' or pv_args.module.casefold() == 'reboot-computer':
                                if pv_args.computer is not None or pv_args.computername is not None:
                                    computername = pv_args.computer if pv_args.computer else pv_args.computername
                                    if temp_powerview:
                                        succeed = temp_powerview.restart_computer(identity=computername, port=445, args=pv_args)
                                    else:
                                        succeed = powerview.restart_computer(identity=computername, port=445, args=pv_args)
                                else:
                                    logging.error('-Computer or -ComputerName is required')
                            elif pv_args.module.casefold() == 'get-netprocess' or pv_args.module.casefold() == 'tasklist':
                                if pv_args.computer is not None or pv_args.computername is not None:
                                    computername = pv_args.computer if pv_args.computer else pv_args.computername
                                    if temp_powerview:
                                        entries = temp_powerview.get_netprocess(identity=computername, port=445, args=pv_args)
                                    else:
                                        entries = powerview.get_netprocess(identity=computername, port=445, args=pv_args)
                                else:
                                    logging.error('-Computer or -ComputerName is required')
                            elif pv_args.module.casefold() == 'stop-netprocess' or pv_args.module.casefold() == 'taskkill':
                                if pv_args.computer is not None or pv_args.computername is not None:
                                    computername = pv_args.computer if pv_args.computer else pv_args.computername
                                    if temp_powerview:
                                        succeed = temp_powerview.stop_netprocess(identity=computername, port=445, args=pv_args)
                                    else:
                                        succeed = powerview.stop_netprocess(identity=computername, port=445, args=pv_args)
                                else:
                                    logging.error('-Computer or -ComputerName is required')

                                if succeed:
                                    logging.info(f'Successfully stopped process {pv_args.pid or pv_args.name} on {computername}')
                                else:
                                    logging.error(f'Failed to stop process {pv_args.pid or pv_args.name} on {computername}')
                            elif pv_args.module.casefold() == 'get-netsession':
                                if pv_args.computer is not None or pv_args.computername is not None:
                                    computername = pv_args.computer if pv_args.computer else pv_args.computername
                                    if temp_powerview:
                                        entries = temp_powerview.get_netsession(identity=computername, port=445, args=pv_args)
                                    else:
                                        entries = powerview.get_netsession(identity=computername, port=445, args=pv_args)
                                else:
                                    logging.error('-Computer or -ComputerName is required')
                            elif pv_args.module.casefold() == 'logoff-session':
                                if pv_args.computer is not None or pv_args.computername is not None:
                                    computername = pv_args.computer if pv_args.computer else pv_args.computername
                                    if temp_powerview:
                                        succeed = temp_powerview.logoff_session(identity=computername, port=445, args=pv_args)
                                    else:
                                        succeed = powerview.logoff_session(identity=computername, port=445, args=pv_args)
                                else:
                                    logging.error('-Computer or -ComputerName is required')
                            elif pv_args.module.casefold() == 'remove-netsession':
                                if pv_args.computer is not None:
                                    if temp_powerview:
                                        succeed = temp_powerview.remove_netsession(computer=pv_args.computer, target_session=pv_args.target_session, args=pv_args)
                                    else:
                                        succeed = powerview.remove_netsession(computer=pv_args.computer, target_session=pv_args.target_session, args=pv_args)
                            elif pv_args.module.casefold() == 'find-localadminaccess':
                                if temp_powerview:
                                    entries = temp_powerview.find_localadminaccess(args=pv_args)
                                else:
                                    entries = powerview.find_localadminaccess(args=pv_args)
                            elif pv_args.module.casefold() == 'invoke-asreproast':
                                if temp_powerview:
                                    entries = temp_powerview.invoke_asreproast(args=pv_args)
                                else:
                                    entries = powerview.invoke_asreproast(args=pv_args)
                            elif pv_args.module.casefold() == 'invoke-kerberoast':
                                if temp_powerview:
                                    entries = temp_powerview.invoke_kerberoast(args=pv_args)
                                else:
                                    entries = powerview.invoke_kerberoast(args=pv_args)
                            elif pv_args.module.casefold() == 'invoke-printerbug':
                                if temp_powerview:
                                    entries = temp_powerview.invoke_printerbug(args=pv_args)
                                else:
                                    entries = powerview.invoke_printerbug(args=pv_args)
                            elif pv_args.module.casefold() == 'invoke-dfscoerce':
                                if temp_powerview:
                                    entries = temp_powerview.invoke_dfscoerce(args=pv_args)
                                else:
                                    entries = powerview.invoke_dfscoerce(args=pv_args)
                            elif pv_args.module.casefold() == 'invoke-messagebox':
                                computername = pv_args.computer if pv_args.computer else pv_args.computername
                                if temp_powerview:
                                    succeed = temp_powerview.invoke_messagebox(identity=computername, args=pv_args)
                                else:
                                    succeed = powerview.invoke_messagebox(identity=computername, args=pv_args)
                            elif pv_args.module.casefold() == 'get-exchangeserver' or pv_args.module.casefold() == 'get-exchange':
                                if temp_powerview:
                                    entries = temp_powerview.get_exchangeserver(args=pv_args)
                                else:
                                    entries = powerview.get_exchangeserver(args=pv_args)
                            elif pv_args.module.casefold() == 'get-exchangemailbox':
                                properties = pv_args.properties if pv_args.properties else None
                                identity = pv_args.identity.strip() if pv_args.identity else None
                                if temp_powerview:
                                    entries = temp_powerview.get_exchangemailbox(identity=identity, properties=properties, args=pv_args)
                                else:
                                    entries = powerview.get_exchangemailbox(identity=identity, properties=properties, args=pv_args)
                            elif pv_args.module.casefold() == 'get-exchangedatabase':
                                properties = pv_args.properties if pv_args.properties else None
                                identity = pv_args.identity.strip() if pv_args.identity else None
                                if temp_powerview:
                                    entries = temp_powerview.get_exchangedatabase(identity=identity, properties=properties, args=pv_args)
                                else:
                                    entries = powerview.get_exchangedatabase(identity=identity, properties=properties, args=pv_args)
                            elif pv_args.module.casefold() == 'unlock-adaccount':
                                if pv_args.identity is not None:
                                    if temp_powerview:
                                        succeed = temp_powerview.unlock_adaccount(identity=pv_args.identity, args=pv_args)
                                    else:
                                        succeed = powerview.unlock_adaccount(identity=pv_args.identity, args=pv_args)
                                else:
                                    logging.error('-Identity flag is required')
                            elif pv_args.module.casefold() == 'add-domaingpo' or pv_args.module.casefold() == 'add-gpo':
                                if pv_args.identity is not None:
                                    if temp_powerview:
                                        succeed = temp_powerview.add_domaingpo(identity=pv_args.identity, description=pv_args.description, basedn=pv_args.basedn, args=pv_args)
                                    else:
                                        succeed = powerview.add_domaingpo(identity=pv_args.identity, description=pv_args.description, basedn=pv_args.basedn, args=pv_args)
                                else:
                                    logging.error('-Identity flag is required')
                            elif pv_args.module.casefold() == 'add-domainou' or pv_args.module.casefold() == 'add-ou':
                                if pv_args.identity is not None:
                                    if temp_powerview:
                                        temp_powerview.add_domainou(identity=pv_args.identity, basedn=pv_args.basedn, args=pv_args)
                                    else:
                                        powerview.add_domainou(identity=pv_args.identity, basedn=pv_args.basedn, args=pv_args)
                                else:
                                    logging.error('-Identity flag is required')
                            elif pv_args.module.casefold() == 'add-netservice':
                                if temp_powerview:
                                    succeed = temp_powerview.add_netservice(computer_name=pv_args.computer, service_name=pv_args.service_name, display_name=pv_args.display_name, binary_path=pv_args.binary_path, service_type=pv_args.service_type, start_type=pv_args.start_type, error_control=pv_args.error_control, service_start_name=pv_args.service_start_name, password=pv_args.password)
                                else:
                                    succeed = powerview.add_netservice(computer_name=pv_args.computer, service_name=pv_args.service_name, display_name=pv_args.display_name, binary_path=pv_args.binary_path, service_type=pv_args.service_type, start_type=pv_args.start_type, error_control=pv_args.error_control, service_start_name=pv_args.service_start_name, password=pv_args.password)
                            elif pv_args.module.casefold() == 'remove-domainou' or pv_args.module.casefold() == 'remove-ou':
                                if pv_args.identity is not None:
                                    if temp_powerview:
                                        temp_powerview.remove_domainou(identity=pv_args.identity, args=pv_args)
                                    else:
                                        powerview.remove_domainou(identity=pv_args.identity, args=pv_args)
                                else:
                                    logging.error('-Identity flag is required')
                            elif pv_args.module.casefold() == 'remove-netservice':
                                if temp_powerview:
                                    succeed = temp_powerview.remove_netservice(computer_name=pv_args.computer, service_name=pv_args.service_name)
                                else:
                                    succeed = powerview.remove_netservice(computer_name=pv_args.computer, service_name=pv_args.service_name)
                            elif pv_args.module.casefold() == 'add-domainobjectacl' or pv_args.module.casefold() == 'add-objectacl':
                                if pv_args.targetidentity is not None and pv_args.principalidentity is not None and pv_args.rights is not None:
                                    if temp_powerview:
                                        temp_powerview.add_domainobjectacl(
                                            targetidentity=pv_args.targetidentity,
                                            principalidentity=pv_args.principalidentity,
                                            rights=pv_args.rights,
                                            rights_guid=pv_args.rights_guid,
                                            ace_type=pv_args.ace_type,
                                            inheritance=pv_args.inheritance
                                        )
                                    else:
                                        powerview.add_domainobjectacl(
                                            targetidentity=pv_args.targetidentity,
                                            principalidentity=pv_args.principalidentity,
                                            rights=pv_args.rights,
                                            rights_guid=pv_args.rights_guid,
                                            ace_type=pv_args.ace_type,
                                            inheritance=pv_args.inheritance
                                        )
                                else:
                                    logging.error('-TargetIdentity , -PrincipalIdentity flags are required')
                            elif pv_args.module.casefold() == 'remove-domainobjectacl' or pv_args.module.casefold() == 'remove-objectacl':
                                if pv_args.targetidentity is not None and pv_args.principalidentity is not None and pv_args.rights is not None:
                                    if temp_powerview:
                                        temp_powerview.remove_domainobjectacl(
                                            targetidentity=pv_args.targetidentity,
                                            principalidentity=pv_args.principalidentity,
                                            rights=pv_args.rights,
                                            rights_guid=pv_args.rights_guid,
                                            ace_type=pv_args.ace_type,
                                            inheritance=pv_args.inheritance
                                        )
                                    else:
                                        powerview.remove_domainobjectacl(
                                            targetidentity=pv_args.targetidentity,
                                            principalidentity=pv_args.principalidentity,
                                            rights=pv_args.rights,
                                            rights_guid=pv_args.rights_guid,
                                            ace_type=pv_args.ace_type,
                                            inheritance=pv_args.inheritance
                                        )
                                else:
                                    logging.error('-TargetIdentity , -PrincipalIdentity flags are required')
                            elif pv_args.module.casefold() == 'add-domaingroupmember' or pv_args.module.casefold() == 'add-groupmember':
                                if pv_args.identity is not None and pv_args.members is not None:
                                    suceed = False
                                    if temp_powerview:
                                        succeed = temp_powerview.add_domaingroupmember(pv_args.identity, pv_args.members, pv_args)
                                    else:
                                        succeed =  powerview.add_domaingroupmember(pv_args.identity, pv_args.members, pv_args)

                                    if succeed:
                                        logging.info(f'User {pv_args.members} successfully added to {pv_args.identity}')
                                else:
                                    logging.error('-Identity and -Members flags required')
                            elif pv_args.module.casefold() == 'remove-domaingroupmember' or pv_args.module.casefold() == 'remove-groupmember':
                                if pv_args.identity is not None and pv_args.members is not None:
                                    suceed = False
                                    if temp_powerview:
                                        succeed = temp_powerview.remove_domaingroupmember(pv_args.identity, pv_args.members, pv_args)
                                    else:
                                        succeed =  powerview.remove_domaingroupmember(pv_args.identity, pv_args.members, pv_args)

                                    if succeed:
                                        logging.info(f'User {pv_args.members} successfully removed from {pv_args.identity}')
                                else:
                                    logging.error('-Identity and -Members flags required')
                            elif pv_args.module.casefold() == 'set-domainobject' or pv_args.module.casefold() == 'set-adobject':
                                if temp_powerview:
                                    succeed = temp_powerview.set_domainobject(pv_args.identity, args=pv_args)
                                else:
                                    succeed = powerview.set_domainobject(pv_args.identity, args=pv_args)
                            elif pv_args.module.casefold() == 'set-domainobjectdn' or pv_args.module.casefold() == 'set-adobjectdn':
                                if pv_args.identity and pv_args.destination_dn:
                                    if temp_powerview:
                                        succeed = temp_powerview.set_domainobjectdn(pv_args.identity, destination_dn=pv_args.destination_dn, args=pv_args)
                                    else:
                                        succeed = powerview.set_domainobjectdn(pv_args.identity, destination_dn=pv_args.destination_dn, args=pv_args)
                                else:
                                    logging.error('-Identity and -DestinationDN flags required')
                            elif pv_args.module.casefold() == 'set-domaindnsrecord':
                                if pv_args.recordname is None or pv_args.recordaddress is None:
                                    logging.error("-RecordName and -RecordAddress flags are required")
                                    continue
                                if temp_powerview:
                                    temp_powerview.set_domaindnsrecord(recordname=pv_args.recordname, recordaddress=pv_args.recordaddress, zonename=pv_args.zonename)
                                else:
                                    powerview.set_domaindnsrecord(recordname=pv_args.recordname, recordaddress=pv_args.recordaddress, zonename=pv_args.zonename)
                            elif pv_args.module.casefold() == 'set-domaincatemplate' or pv_args.module.casefold() == 'set-catemplate':
                                if pv_args.identity and (pv_args.clear or pv_args.set or pv_args.append):
                                    if temp_powerview:
                                        temp_powerview.set_domaincatemplate(pv_args.identity, pv_args)
                                    else:
                                        powerview.set_domaincatemplate(pv_args.identity, pv_args)
                                else:
                                    logging.error('-Identity and [-Clear][-Set|-Append] flags required')
                            elif pv_args.module.casefold() == 'set-domainuserpassword':
                                if pv_args.identity and pv_args.accountpassword:
                                    succeed = False
                                    if temp_powerview:
                                        succeed = temp_powerview.set_domainuserpassword(pv_args.identity, pv_args.accountpassword, oldpassword=pv_args.oldpassword, args=pv_args)
                                    else:
                                        succeed = powerview.set_domainuserpassword(pv_args.identity, pv_args.accountpassword, oldpassword=pv_args.oldpassword, args=pv_args)

                                    if succeed:
                                        logging.info(f'Password changed for {pv_args.identity}')
                                    else:
                                        logging.error(f'Failed password change attempt for {pv_args.identity}')
                                else:
                                    logging.error('-Identity and -AccountPassword flags are required')
                            elif pv_args.module.casefold() == 'set-domaincomputerpassword':
                                if pv_args.identity and pv_args.accountpassword:
                                    succeed = False
                                    if temp_powerview:
                                        succeed = temp_powerview.set_domaincomputerpassword(pv_args.identity, pv_args.accountpassword, oldpassword=pv_args.oldpassword, args=pv_args)
                                    else:
                                        succeed = powerview.set_domaincomputerpassword(pv_args.identity, pv_args.accountpassword, oldpassword=pv_args.oldpassword, args=pv_args)

                                    if succeed:
                                        logging.info(f'Password changed for {pv_args.identity}')
                                    else:
                                        logging.error(f'Failed password change attempt for {pv_args.identity}')
                                else:
                                    logging.error('-Identity and -AccountPassword flags are required')
                            elif pv_args.module.casefold() == 'set-domainrbcd' or pv_args.module.casefold() == 'set-rbcd':
                                if pv_args.delegatefrom is not None and pv_args.identity is not None:
                                    if temp_powerview:
                                        temp_powerview.set_domainrbcd(pv_args.identity, pv_args.delegatefrom, args=pv_args)
                                    else:
                                        powerview.set_domainrbcd(pv_args.identity, pv_args.delegatefrom, args=pv_args)
                                else:
                                    logging.error('-Identity and -DelegateFrom flags are required')
                            elif pv_args.module.casefold() == 'set-domainobjectowner' or pv_args.module.casefold() == 'set-objectowner':
                                if pv_args.targetidentity is not None and pv_args.principalidentity is not None:
                                    if temp_powerview:
                                        temp_powerview.set_domainobjectowner(pv_args.targetidentity, pv_args.principalidentity, args=pv_args)
                                    else:
                                        powerview.set_domainobjectowner(pv_args.targetidentity, pv_args.principalidentity, args=pv_args)
                                else:
                                    logging.error('-TargetIdentity and -PrincipalIdentity flags are required')
                            elif pv_args.module.casefold() == 'add-domaincomputer' or pv_args.module.casefold() == 'add-adcomputer':
                                if pv_args.computername is not None:
                                    if pv_args.computerpass is None:
                                        pv_args.computerpass = ''.join(random.choice(list(string.ascii_letters + string.digits + "!@#$%^&*()")) for _ in range(12))
                                    if temp_powerview:
                                        temp_powerview.add_domaincomputer(pv_args.computername, pv_args.computerpass, basedn=pv_args.basedn)
                                    else:
                                        powerview.add_domaincomputer(pv_args.computername, pv_args.computerpass, basedn=pv_args.basedn)
                                else:
                                    logging.error(f'-ComputerName and -ComputerPass are required')
                            elif pv_args.module.casefold() == 'add-domaingmsa' or pv_args.module.casefold() == 'add-gmsa':
                                if temp_powerview:
                                    temp_powerview.add_domaingmsa(args=pv_args)
                                else:
                                    powerview.add_domaingmsa(args=pv_args)
                            elif pv_args.module.casefold() == 'add-domaindmsa' or pv_args.module.casefold() == 'add-dmsa':
                                if temp_powerview:
                                    temp_powerview.add_domaindmsa(args=pv_args)
                                else:
                                    powerview.add_domaindmsa(args=pv_args)
                            elif pv_args.module.casefold() == 'add-domaindnsrecord':
                                if pv_args.recordname is None or pv_args.recordaddress is None:
                                    logging.error("-RecordName and -RecordAddress flags are required")
                                    continue
                                if temp_powerview:
                                    temp_powerview.add_domaindnsrecord(recordname=pv_args.recordname, recordaddress=pv_args.recordaddress, zonename=pv_args.zonename)
                                else:
                                    powerview.add_domaindnsrecord(recordname=pv_args.recordname, recordaddress=pv_args.recordaddress, zonename=pv_args.zonename)
                            elif pv_args.module.casefold() == 'add-domainuser' or pv_args.module.casefold() == 'add-aduser':
                                if temp_powerview:
                                    temp_powerview.add_domainuser(pv_args.username, pv_args.userpass, args=pv_args)
                                else:
                                    powerview.add_domainuser(pv_args.username, pv_args.userpass, args=pv_args)
                            elif pv_args.module.casefold() == 'add-domaingroup' or pv_args.module.casefold() == 'add-adgroup':
                                if pv_args.identity is not None:
                                    if temp_powerview:
                                        temp_powerview.add_domaingroup(pv_args.identity, basedn=pv_args.basedn, args=pv_args)
                                    else:
                                        powerview.add_domaingroup(pv_args.identity, basedn=pv_args.basedn, args=pv_args)
                                else:
                                    logging.error('-Name flag is required')
                            elif pv_args.module.casefold() == 'remove-domainobject' or pv_args.module.casefold() == 'remove-adobject':
                                if pv_args.identity:
                                    identity = pv_args.identity.strip()
                                    if temp_powerview:
                                        temp_powerview.remove_domainobject(identity, args=pv_args)
                                    else:
                                        powerview.remove_domainobject(identity, args=pv_args)
                                else:
                                    logging.error("-Identity flag is required")
                            elif pv_args.module.casefold() == 'remove-domaindmsa' or pv_args.module.casefold() == 'remove-dmsa':
                                if temp_powerview:
                                    temp_powerview.remove_domaindmsa(args=pv_args)
                                else:
                                    powerview.remove_domaindmsa(args=pv_args)
                            elif pv_args.module.casefold() == 'remove-domaingmsa' or pv_args.module.casefold() == 'remove-gmsa':
                                if temp_powerview:
                                    temp_powerview.remove_domaingmsa(args=pv_args)
                                else:
                                    powerview.remove_domaingmsa(args=pv_args)
                            elif pv_args.module.casefold() == 'remove-domainuser' or pv_args.module.casefold() == 'remove-aduser':
                                if pv_args.identity:
                                    if temp_powerview:
                                        temp_powerview.remove_domainuser(pv_args.identity)
                                    else:
                                        powerview.remove_domainuser(pv_args.identity)
                                else:
                                    logging.error("-Identity is required")
                            elif pv_args.module.casefold() == 'remove-domaindnsrecord':
                                if not pv_args.recordname:
                                    logging.error("-RecordName flag is required")
                                    continue
                                if temp_powerview:
                                    temp_powerview.remove_domaindnsrecord(recordname=pv_args.recordname, zonename=pv_args.zonename)
                                else:
                                    powerview.remove_domaindnsrecord(recordname=pv_args.recordname, zonename=pv_args.zonename)
                            elif pv_args.module.casefold() == 'disable-domaindnsrecord':
                                if not pv_args.recordname:
                                    logging.error("-RecordName flag is required")
                                    continue
                                if temp_powerview:
                                    temp_powerview.disable_domaindnsrecord(recordname=pv_args.recordname, zonename=pv_args.zonename)
                                else:
                                    powerview.disable_domaindnsrecord(recordname=pv_args.recordname, zonename=pv_args.zonename)
                            elif pv_args.module.casefold() == 'restore-domainobject' or pv_args.module.casefold() == 'restore-adobject':
                                if temp_powerview:
                                    temp_powerview.restore_domainobject(args=pv_args)
                                else:
                                    powerview.restore_domainobject(args=pv_args)
                            elif pv_args.module.casefold() == 'remove-domaincomputer' or pv_args.module.casefold() == 'remove-adcomputer':
                                if pv_args.computername is not None:
                                    if temp_powerview:
                                        temp_powerview.remove_domaincomputer(pv_args.computername, args=pv_args)
                                    else:
                                        powerview.remove_domaincomputer(pv_args.computername, args=pv_args)
                                else:
                                    logging.error('-ComputerName is required')
                            elif pv_args.module.casefold() == 'add-gplink':
                                if pv_args.guid is not None and pv_args.targetidentity is not None:
                                    if temp_powerview:
                                        powerview.add_gplink(guid=pv_args.guid, targetidentity=pv_args.targetidentity, link_enabled=pv_args.link_enabled, enforced=pv_args.enforced, args=pv_args)
                                    else:
                                        powerview.add_gplink(guid=pv_args.guid, targetidentity=pv_args.targetidentity, link_enabled=pv_args.link_enabled, enforced=pv_args.enforced, args=pv_args)
                                else:
                                    logging.error("-GUID and -TargetIdentity flags are required")
                            elif pv_args.module.casefold() == 'remove-gplink':
                                if pv_args.guid is not None and pv_args.targetidentity is not None:
                                    if temp_powerview:
                                        powerview.remove_gplink(guid=pv_args.guid, targetidentity=pv_args.targetidentity, args=pv_args)
                                    else:
                                        powerview.remove_gplink(guid=pv_args.guid, targetidentity=pv_args.targetidentity, args=pv_args)
                                else:
                                    logging.error("-GUID and -TargetIdentity flags are required")
                            elif pv_args.module.casefold() == 'get_pool_stats':
                                if temp_powerview:
                                    stats = temp_powerview.conn.get_pool_stats()
                                else:
                                    stats = powerview.conn.get_pool_stats()
                                
                                FORMATTER.format_pool_stats(stats)
                            elif pv_args.module.casefold() == 'history':
                                hist = get_shell_history(pv_args.last, pv_args.unique)
                                for index, item in list(enumerate(hist,1))[::-1]:
                                    if pv_args.noNumber:
                                        bol = f""
                                    else:
                                        bol = f"[{index}] "
                                    print(f"{bol}{item}")
                            elif pv_args.module.casefold() == 'clear':
                                clear_screen()
                            elif pv_args.module.casefold() == 'whoami':
                                print(powerview.conn.who_am_i())
                            elif pv_args.module.casefold() == 'exit':
                                if args.mcp and hasattr(powerview, 'mcp_server') and powerview.mcp_server.get_status():
                                    powerview.mcp_server.stop()
                                log_handler.save_history()
                                sys.exit(0)

                            if entries:
                                if pv_args.outfile:
                                    if os.path.exists(pv_args.outfile):
                                        logging.error("%s exists "%(pv_args.outfile))
                                        continue

                                formatter = FORMATTER(pv_args, args.use_kerberos)
                                if hasattr(pv_args, 'where') and pv_args.where is not None:
                                    entries = formatter.alter_entries(entries,pv_args.where)

                                if hasattr(pv_args, 'sort_by') and pv_args.sort_by is not None:
                                    entries = formatter.sort_entries(entries,pv_args.sort_by)

                                if entries is None:
                                    logging.error(f'Key not available')
                                else:
                                    if hasattr(pv_args, "count") and pv_args.count:
                                        formatter.count(entries)
                                    elif hasattr(pv_args, "tableview") and pv_args.tableview:
                                        formatter.table_view(entries)
                                    elif hasattr(pv_args, "select") and pv_args.select is not None:
                                        if isinstance(pv_args.select, int):
                                            formatter.print_index(entries)
                                        else:
                                            formatter.print_select(entries)
                                    else:
                                        if isinstance(entries, dict) and entries.get("headers"):
                                            formatter.print_table(entries["rows"], entries["headers"])
                                        else:
                                            formatter.print(entries)

                            # After displaying results, check if they came from cache
                            if entries and len(entries) > 0:
                                first_entry = entries[0]
                                using_cache = isinstance(first_entry, dict) and first_entry.get('from_cache', False)

                            temp_powerview = None
                            current_target_domain = None
                            conn.set_ldap_address(init_ldap_address)
                            conn.set_targetDomain(None)
                        except ldap3.core.exceptions.LDAPInvalidFilterError as e:
                            logging.error(str(e))
                        except ldap3.core.exceptions.LDAPAttributeError as e:
                            logging.error(str(e))
                        except ldap3.core.exceptions.LDAPSocketSendError as e:
                            logging.error(str(e))
                            conn.reset_connection()
                        except ldap3.core.exceptions.LDAPSocketReceiveError as e:
                            logging.error(str(e))
                            conn.reset_connection()
            except KeyboardInterrupt:
                if args.mcp and hasattr(powerview, 'mcp_server') and powerview.mcp_server.get_status():
                    powerview.mcp_server.stop()
                log_handler.save_history()
                print()
            except EOFError:
                if args.mcp and hasattr(powerview, 'mcp_server') and powerview.mcp_server.get_status():
                    powerview.mcp_server.stop()
                log_handler.save_history()
                print("Exiting...")
                conn.close()
                sys.exit(0)
            except (ldap3.core.exceptions.LDAPSocketSendError, 
                    ldap3.core.exceptions.LDAPSocketReceiveError) as e:
                logging.info(f"LDAP Socket Error: {str(e)}")
                conn.reset_connection()
            except ldap3.core.exceptions.LDAPSessionTerminatedByServerError:
                logging.warning("Server connection terminated. Trying to reconnect")
                conn.reset_connection()
                continue
            except ldap3.core.exceptions.LDAPInvalidDnError as e:
                logging.error(f"LDAPInvalidDnError: {str(e)}")
                continue
            except Exception as e:
                if args.stack_trace:
                    log_handler.save_history()
                    raise
                else:
                    logging.error(str(e))

            if args.query:
                conn.close()
                sys.exit(0)

    except ldap3.core.exceptions.LDAPSocketOpenError as e:
        print(str(e))
    except ldap3.core.exceptions.LDAPBindError as e:
        print(str(e))
    except Exception as e:
        if args.stack_trace:
            log_handler.save_history()
            raise
        else:
            logging.error(str(e))

if __name__ == '__main__':
    main()
